// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Decimate.h"
#include <sys/timeb.h>

// Use the standard c++ libs
using namespace std;

Decimate::Decimate() {}

/*!
  The purpose of this MQP is to decimate, so this is the function
  that decimates a scene. It is passed two parameters. To decimate
  we need a scene, from which we remove vertices, and a number of
  vertices to remove.
*/
bool Decimate::remove(Scene * scene, int maxNumRemoved) {
  timeb timeBefore, timeAfter;
  int numRemoved = 0;
  int numFailed = 0;

  if (maxNumRemoved <= 0) {
    cout << "Scene has been fully decimated." << endl;
    return true;
  }

  cout << "Removing " << maxNumRemoved << " vertices." << endl;

  // Set up to time the decimation
  ftime( &timeBefore );

  // stop when we have removed as many as desired
  while (numRemoved != maxNumRemoved) {
    // the best candidate is the Vertex we will try to remove
    Vertex * bestCandidate = 0;

    // we need to know what Model the Vertex came from
    int model_num = 0;

    // for each model we need to find the best candidate vertex
    // in the entire scene. this gives an even distribution of
    // removals.

    // the best candidate is defined as the vertex with the
    // smallest distance to the plane AND a non-corner

    // a non-corner is any vertex that has not been flagged as
    // a corner in the initial Scene analysis. (we never remove
    // corners!)

    // this is much easier if we fix auto-sorting on distances,
    // since we would only need to find the first non-corner.
    // <set> is supposed to be automatically sorted
    VS * vertices;
    VSI v;
    for (int i = 0; i != scene->numModels(); i++) {
      // iterate through the vertex list in this model
      // to find the best candidate in this model
      vertices = scene->m(i)->getVertexList();
      for (v = vertices->begin(); v != vertices->end(); v++) {
        if ( !(*v)->isCorner()) {
          if (!bestCandidate || bestCandidate->getDistance() > (*v)->getDistance()) {
            bestCandidate = (*v);
            model_num = i;
          }
        }
      }
    }  // end of for each model

    // there were no candidates in the entire scene
    if (!bestCandidate) break;

    // then we remove the best candidate from the Model (hence the Scene)
    if (removeVertex(scene->m(model_num), bestCandidate)) {
      if (numRemoved % 50 == 0)
        cout << endl << "  " << numRemoved << "\t--> ";
      cout << ".";
      cout.flush();

      numRemoved++;
      // remove any models with less than a triangle (it doesn't render)
      if (scene->m(model_num)->numVertices() < 3) {
        numRemoved += scene->m(model_num)->numVertices();
        scene->removeModel(scene->m(model_num));
      }
    }
    else {
      numFailed++;
    }

    if (numRemoved + numFailed >= maxNumRemoved) break;
  }  // end of while

  // Finish up timing the decimation
  ftime( &timeAfter );
  float timeTaken = ( timeAfter.time - timeBefore.time ) +
                    ( float )( timeAfter.millitm - timeBefore.millitm )
                    / 1000;

  cout << "\n\nRemoved:\t" << numRemoved << endl;
  cout << "Failed:\t\t" << numFailed << endl;
  cout << "Time: " << timeTaken << "s" << endl;

  // Return with error notification
  if (numRemoved == maxNumRemoved) return true;
  else return false;
}

/*
  This version of removeVertex() uses replacement. But chooses a better head.

  There are five sets of data that we must concern ourselves with when we delete.
  two data sets of triangles and three data sets of vertices
  1. one set of Triangle in each neighbor Vertex
  2. one set of Triangle in each Model
  3. one set of Vertex in each Model
  4. one set of Vertex in each neighbor Vertex
  5. one set of Vertex in each Triangle

  so in order to delete a vertex and triangle from memory these data sets
  must be robust and without error. without a Triangle having a deleted Vertex,
  a Vertex having a deleted Triangle, a Vertex having a deleted neighboring Vertex,
  or a Model with either a deleted Vertex, or Triangle.
*/
/*
bool Decimate::removeVertex(Model * model, Vertex * vertex_to_remove) {
  // 1. isolate neighbors (easy)
  set <Vertex *> * neighbors = vertex_to_remove->getNeighbors();

  if (neighbors->size() == 0) {  // no neighbors -> we remove and delete
    model->removeVertex(vertex_to_remove);

    delete vertex_to_remove;
    return true;
  }
  else if (neighbors->size() == 1) {  // one neighbor -> we remove and delete
    set <Vertex *>::iterator popper = neighbors->begin();
    //grab the first neighbor
    Vertex * other = (*popper);

    other->removeNeighbor(vertex_to_remove);

    model->removeVertex(vertex_to_remove);

    // free the memory
    delete vertex_to_remove;

    return true;
  }
  // this is a special case, we can deal with this later (edge case only has two neighbors)
  else if (neighbors->size() == 2)  {
    // we need to keep track of the triangles that need to be deleted, so we can free up memory
    vector <Triangle *> triangles_to_delete;

    set <Vertex *>::iterator popper = neighbors->begin();
    //grab the first neighbor
    Vertex * first = (*popper);
    popper++;
    //grab the second neighbor
    Vertex * second = (*popper);

    // need to remove the triangles associated
    // and create a new one
    set <Triangle *> * old_triangles = vertex_to_remove->getTriangles();

    for (set <Triangle *>::iterator tri = old_triangles->begin();
         tri != old_triangles->end();
         tri++) {
      model->removeTriangle(*tri);
      triangles_to_delete.push_back(*tri);
    }

    first->replace(vertex_to_remove, second);
    second->replace(vertex_to_remove, first);

    model->removeVertex(vertex_to_remove);

    // free the memory
    int tri_size = triangles_to_delete.size();
    for (int i = 0; i != tri_size; i++) {
      delete triangles_to_delete[i];
    }

    delete vertex_to_remove;

    return true;
  }


  typedef multimap <Vertex *, Vertex *>::iterator MI;
  set <Triangle *> triangle_list;  // the list of Triangles that we add to the model later
  multimap <Vertex *, Vertex *> temp_neighbors;  // the list of neighbors mapped to their neighbors


  // we need a copy of neighbors so we don't mess with the original
  set <Vertex *> ncopy;
  set <Vertex *>::iterator n;
  for (n = neighbors->begin();
       n != neighbors->end();
       n++) {
    ncopy.insert(*n);
  }


  // this creates a mapping between a vertex in the neighbor list
  // with another neighbor in the list if they are neighbors
  // basically it creates a temporary bond between known neighbors
//  cout << "before temp_neighbors:" << endl;
  for (set <Vertex *>::iterator neigh = ncopy.begin();
       neigh != ncopy.end();
       neigh++) {
    for (set <Vertex *>::iterator in = ncopy.begin();
         in != ncopy.end();
         in++) {
      if ((*neigh)->contains(*in)) {
        temp_neighbors.insert(make_pair(*neigh, *in));
//        cout << "\tinsert neighbors: (" << (*neigh)->getID() << ", " << (*in)->getID() << ")" << endl;
      }
    }
  }

//  cout << "temp_neighbors:" << endl;
//  for (MI t_iter = temp_neighbors.begin(); t_iter != temp_neighbors.end(); t_iter++) {
//    cout << "\tneighbors: (" << t_iter->first->getID() << ", " << t_iter->second->getID() << ")" << endl;
//  }


  // determines if we have exhausted all possibilities
  bool exhausted = false;
  set <Vertex *> choosen_already;

  while (!exhausted && ncopy.size() != 3) {  // if we have three vertex left we stop
    Vertex * choosen_n, * middle_n, * link_n;
    pair <MI, MI> range;
    MI iter;

    // get choosen_n (the first neighbor with two shared neighbors)
    for (n = ncopy.begin(); n != ncopy.end(); n++) {
      // check if we have used choosen_n before


      range = temp_neighbors.equal_range(*n);

      int count = 0;  // the number of neighbors
      // count how many neighbors it has
      for (iter = range.first; iter != range.second; iter++) {
        count++;
	  }

      // if it has two neighbors we use it
      if (count == 2 && choosen_already.find(*n) == choosen_already.end()) {
        choosen_n = *n;
        choosen_already.insert(*n);
        break;
      }
    }

    // get the range of neighbors of choosen_n (all possible middle_n's)
    range = temp_neighbors.equal_range(choosen_n);
    bool link_created = false;

    // for each possible middle_n (we checked above, there is only two)
    for (iter = range.first; iter != range.second; iter++) {
      middle_n = iter->second;
      pair <MI, MI> middle_n_range = temp_neighbors.equal_range(middle_n);

      set <Vertex *> link_n_list;
      for (MI middle_n_iter = middle_n_range.first; middle_n_iter != middle_n_range.second; middle_n_iter++) {
        link_n_list.insert(middle_n_iter->second);
//        cout << "link_n_list.insert: " << middle_n_iter->second->getID() << endl;
	  }


      // get link_n (one of middle_n's neighbors)
      if (link_n_list.size() == 2 || link_n_list.size() == 3) {  // in this case link_n is not choosen_n, but the other neighbor
        for (set <Vertex *>::iterator i = link_n_list.begin(); i != link_n_list.end(); i++) {
          pair <MI, MI> temp_range = temp_neighbors.equal_range(*i);
          int count = 0;

          for (MI temp_iter = temp_range.first; temp_iter != temp_range.second; temp_iter++) {
            count++;
		  }

          if ((*i) != choosen_n && count != 3)
            link_n = *i;
		}
	  }
      else { cout << "wtf: link_n_list.size() = " << link_n_list.size() << endl; continue; }

//      cout << "choosen_n = " << choosen_n->getID() << endl;
//      cout << "middle_n = " << middle_n->getID() << endl;
//      cout << "link_n = " << link_n->getID() << endl;
//      cout << "vertex_to_remove = " << vertex_to_remove->getID() << endl;

      // check for concavities
      if (can_create_link(choosen_n, middle_n, link_n, vertex_to_remove)) {
//        cout << "create the link!" << endl;
        Triangle * t;

        if (ordered(choosen_n, middle_n, link_n, model->isClockwise()))
          t = new Triangle(choosen_n, middle_n, link_n);
        else
          t = new Triangle(link_n, middle_n, choosen_n);

        triangle_list.insert(t);

        temp_neighbors.erase(middle_n);

        vector <MI> temp_set;

        // remove the neighbor pairs associated with middle_n in temp_neighbors (they no longer apply to the triangulation)
        for (MI temp_iter = temp_neighbors.begin(); temp_iter != temp_neighbors.end(); temp_iter++) {
          if (temp_iter->second == middle_n)
            temp_set.push_back(temp_iter);
		}

        for (vector <MI>::iterator set_iter = temp_set.begin(); set_iter != temp_set.end(); set_iter++)
          temp_neighbors.erase(*set_iter);

        // then remove the middle_n from the neighbors list (it has been isolated)
        ncopy.erase(middle_n);

        // now associate the new neighbors
        temp_neighbors.insert(make_pair(choosen_n, link_n));
        temp_neighbors.insert(make_pair(link_n, choosen_n));
//        cout << "\tadded neighbors: (" << choosen_n->getID() << ", " << link_n->getID() << ")" << endl;
//        cout << "\tadded neighbors: (" << link_n->getID() << ", " << choosen_n->getID() << ")" << endl;

        link_created = true;
        break;  // break from for each possible middle_n (it was a sucess)
	  } else cout << "FAILED: can_create_link!" << endl;
    }  // end of for each possible middle_n

    if (!link_created)  // link was not created (ran out of possible middle_n's)
      if (choosen_already.size() == neighbors->size()) exhausted = true;

	  if (ncopy.size() == 3) {  // if there are only three neighbors left we need to create a Triangle with them
      set <Vertex *>::iterator vert = ncopy.begin();
      Vertex * v0 = (*vert);
      vert++;
      Vertex * v1 = (*vert);
      vert++;
      Vertex * v2 = (*vert);

      if (ordered(v0, v1, v2, model->isClockwise()))
        triangle_list.insert(new Triangle(v0, v1, v2));
      else
        triangle_list.insert(new Triangle(v2, v1, v0));
    }
  }  // end of while (!exhausted && ncopy.size() != 3)

  //cout << "\ntemp_neighbors:" << endl;
  //for (MI tmp_iter = temp_neighbors.begin(); tmp_iter != temp_neighbors.end(); tmp_iter++) {
    //cout << "\tneighbor: (" << tmp_iter->first->getID() << ", " << tmp_iter->second->getID() << ")" << endl;
  //}


  // have we exhausted all possibilities?
  if (exhausted) return false;

  // remove all triangles from model
  set <Triangle *> * triangles_to_remove = vertex_to_remove->getTriangles();
  set <Triangle *>::iterator t;
  set <Vertex *>::iterator v;
  for (t = triangles_to_remove->begin(); t != triangles_to_remove->end(); t++) {
    model->removeTriangle(*t);

    // remove all triangles from the neighbors
    for (v = neighbors->begin(); v != neighbors->end(); v++) {
      (*v)->removeTriangle(*t);
    }
  }

  // remove vertex_to_remove from model
  model->removeVertex(vertex_to_remove);

  set <Vertex *> temp_n = *neighbors;
  // remove vertex_to_remove from neighbors
  for (v = neighbors->begin(); v != neighbors->end(); v++) {
    temp_n.insert(*v);
  }

  for (v = temp_n.begin(); v != temp_n.end(); v++) {
    (*v)->removeNeighbor(vertex_to_remove);
  }

  // delete all triangles
  // delete vertex_to_remove

  // create new triangles
  for (set <Triangle *>::iterator tri = triangle_list.begin(); tri != triangle_list.end(); tri++) {
    model->addTriangle(*tri);
    model->addVertex((*tri)->v0());
    model->addVertex((*tri)->v1());
    model->addVertex((*tri)->v2());
  }
}
*/

bool Decimate::ordered(Vertex * v0, Vertex * v1, Vertex * v2, bool clockwise) {
  return true;

}


bool Decimate::can_create_link(Vertex * c_n, Vertex * m_n, Vertex * l_n, Vertex * v_t_r) {
  Vector v0 = Vector(v_t_r->x() - m_n->x(), v_t_r->y() - m_n->y(), v_t_r->z() - m_n->z());
  Vector v1 = Vector(c_n->x() - m_n->x(), c_n->y() - m_n->y(), c_n->z() - m_n->z());
  Vector v2 = Vector(l_n->x() - m_n->x(), l_n->y() - m_n->y(), l_n->z() - m_n->z());

  float angle = acos(dot(v1, v0) / (v1.magnitude() * v0.magnitude())) + acos(dot(v0, v2) / (v0.magnitude() * v2.magnitude()));

  cout << "angle = " << angle << endl;
  if (angle > 3.124f)  // about 179
    return false;

  return true;
}

bool Decimate::removeVertex(Model * model, Vertex * vertex_to_remove) {
  // 1. isolate neighbors (easy)
  set <Vertex *> * neighbors = vertex_to_remove->getNeighbors();

  if (neighbors->size() == 0) {  // no neighbors -> we remove and delete
    model->removeVertex(vertex_to_remove);

    delete vertex_to_remove;
    return true;
  }
  else if (neighbors->size() == 1) {  // one neighbor -> we remove and delete
    set <Vertex *>::iterator popper = neighbors->begin();
    //grab the first neighbor
    Vertex * other = (*popper);

    other->removeNeighbor(vertex_to_remove);

    model->removeVertex(vertex_to_remove);

    // free the memory
    delete vertex_to_remove;

    return true;
  }
  // this is a special case, we can deal with this later (edge case only has two neighbors)
  else if (neighbors->size() == 2)  {
    // we need to keep track of the triangles that need to be deleted, so we can free up memory
    vector <Triangle *> triangles_to_delete;

    set <Vertex *>::iterator popper = neighbors->begin();
    //grab the first neighbor
    Vertex * first = (*popper);
    popper++;
    //grab the second neighbor
    Vertex * second = (*popper);

    // need to remove the triangles associated
    // and create a new one
    set <Triangle *> * old_triangles = vertex_to_remove->getTriangles();

    for (set <Triangle *>::iterator tri = old_triangles->begin();
         tri != old_triangles->end();
         tri++) {
      model->removeTriangle(*tri);
      triangles_to_delete.push_back(*tri);
    }

    first->replace(vertex_to_remove, second);
    second->replace(vertex_to_remove, first);

    model->removeVertex(vertex_to_remove);

    // free the memory
    int tri_size = triangles_to_delete.size();
    for (int i = 0; i != tri_size; i++) {
      delete triangles_to_delete[i];
    }

    delete vertex_to_remove;

    return true;
  }

  // the variable that determines if we found a suitable head
  bool found = false;


  // CHOOSES THE LIST OF POSSIBLE HEADS
  // set of vertices that are possible heads
  vector <Vertex *> heads;
  map <Vertex *, float> distances_to_vertex;
  for (set <Vertex *>::iterator v = neighbors->begin(); v != neighbors->end(); v++) {
    Vector vect = Vector((*v)->x() - vertex_to_remove->x(), (*v)->y() - vertex_to_remove->y(), (*v)->z() - vertex_to_remove->z());

    if (heads.empty()) {
      heads.insert(heads.end(),*v);    // directly insert the first vertex
    }
    else {    // order all vertices after the first one
      int size = heads.size();
      for (map <Vertex *, float>::iterator x = distances_to_vertex.begin(); x != distances_to_vertex.end(); x++) {
        for (vector <Vertex *>::iterator point_to_vertex = heads.begin(); point_to_vertex != heads.end(); point_to_vertex++) {
          if (x->first == (*point_to_vertex)) {    // found the right vertex
            if (vect.magnitude() < x->second) {    // is it less distance to the point than the iterator
              // the possible heads are the neighbors, but we
	      // directly alter the set 'neighbors' so we have to copy them
              heads.insert(point_to_vertex, *v);
	      break;
	    }
	  }
	}
      }
      // if heads is the same size as it was then we need to
      // insert the vertex before the end
      if ((unsigned)size == (unsigned)heads.size())
	heads.insert(heads.end(), *v);
    }

    distances_to_vertex[*v] = vect.magnitude();
  }

  // we need a set to keep track of triangles that need to be removed
  set <Triangle *> collapsed_triangles;

  // we need to keep track of the triangles that need to be deleted, so we can free up memory
  vector <Triangle *> triangles_to_delete;

  // we need a set to hold the unmodified triangles
  set <Triangle *> old_triangles = *vertex_to_remove->getTriangles();

  // we need an iterator to go through the entire list of possible heads
  vector <Vertex *>::iterator possible_head = heads.begin();
  while (!found && possible_head != heads.end()) {
    // 2. pick head (easy)
    Vertex * head = *(possible_head++);

    // 3. modify triangles to represent new triangle structure (intermediate)
    // we need an iterator for inside the for loop
    set <Triangle *>::iterator tri;

    // 3a. replace vertex_to_remove with new head (intermediate)
    // if it fails to replace them then the triangle has collapsed
    collapsed_triangles.clear();
    triangles_to_delete.clear();
    for (tri = old_triangles.begin(); tri != old_triangles.end(); tri++) {
      if (!(*tri)->replace(vertex_to_remove, head)) {
        collapsed_triangles.insert(*tri);
        triangles_to_delete.push_back(*tri);
      }
    }
    
    // 3b. replace vertex_to_remove with new head in neighbors (should not be needed, but somewhere in 3a. it doesn't replace them all)
    set <Vertex *> ncopy = *neighbors;
    for (set <Vertex *>::iterator iter = ncopy.begin(); iter != ncopy.end(); iter++) {
      (*iter)->replace(vertex_to_remove, head);
    }

    // check for any overlapping triangles
    // if this returns false we need to pick another head, because it overlaps
    found = true;  // check_overlap(&heads, &old_triangles);
    //if (!found) {
      // we need to clean-up
      //for (tri = old_triangles.begin(); tri != old_triangles.end(); tri++) {
        //if ((*tri)->replaced())    // if replaced before
          //(*tri)->replace(head, vertex_to_remove);    // replace it back to original state
      //}

      //cout << "ERROR: no suitable head was found... retrying" << endl;
    //}
  }

  if (!found) return false;

  // an iterator for use outside of the for loop
  set <Triangle *>::iterator t;

  // 3c. if the replacement was not sucessful (from above) then the
  //     triangle has collapsed, and now we need to erase
  //     the triangle from the list (easy). these are the triangles
  //     that have collapsed from the join
  for (t = collapsed_triangles.begin(); t != collapsed_triangles.end(); t++) {
    model->removeTriangle(*t);
  }

  // 3d. erase the vertex
  model->removeVertex(vertex_to_remove);
  if (vertex_to_remove->getNeighbors()->size() != 0) cout << "neighbors = " << vertex_to_remove->getNeighbors()->size() << endl;
  // free the memory
  int tri_size = triangles_to_delete.size();
  for (int i = 0; i != tri_size; i++) {
    delete triangles_to_delete[i];
  }

  delete vertex_to_remove;

  return true;
}

bool Decimate::check_overlap(vector <Vertex *> * verts, set <Triangle *> * tris) {
  for (vector <Vertex *>::iterator vert = verts->begin(); vert != verts->end(); vert++) {
    for (set <Triangle *>::iterator tri = tris->begin(); tri != tris->end(); tri++) {
      if (vertex_contained(*vert, *tri)) return false;
    }
  }

  return true;
}

bool Decimate::vertex_contained(Vertex * v, Triangle * t) {
  // i heard somewhere that if you sum the angles of vectors you get 2pi
  Vector v0 = Vector(t->v0()->x() - v->x(), t->v0()->y() - v->y(), t->v0()->z() -  v->z());
  Vector v1 = Vector(t->v1()->x() - v->x(), t->v1()->y() - v->y(), t->v1()->z() -  v->z());
  Vector v2 = Vector(t->v2()->x() - v->x(), t->v2()->y() - v->y(), t->v2()->z() -  v->z());

  float x0 = acos(dot(v0, v1) / (v0.magnitude() * v1.magnitude()));
  float x1 = acos(dot(v1, v2) / (v1.magnitude() * v2.magnitude()));
  float x2 = acos(dot(v2, v0) / (v2.magnitude() * v0.magnitude()));


  float sum_of_angles = x0 + x1 + x2;

  if (fabs(sum_of_angles - 2 * 3.1415926) < 0.1) return true;

  return false;
}

