// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Model.h"

using namespace std;

Model::Model(vector <Vertex *> * v,
	     vector <GLint> * vi,
	     vector <Color *> * c,
	     bool order) {
  clockwise = order;
  configureVectors(v, vi, c);
}


void Model::addTriangle(Triangle * t) { triangles.insert(triangles.end(), t); }
void Model::removeTriangle(Triangle * t) {
  set <Triangle *>::iterator i = triangles.find(t);
  if (i != triangles.end()) {
    triangles.erase(i);
  }
}


void Model::addVertex(Vertex * v) { vertices.insert(vertices.end(), v); }
void Model::removeVertex(Vertex * v) {
  VSI i = vertices.find(v);
  if (i != vertices.end()) {
    vertices.erase(i);
  }
}

bool Model::isClockwise() { return clockwise; }

/*
  Computes the normal given three Vertex
*/
Normal * Model::computeNormals(Vertex * x0, Vertex * x1, Vertex * x2) {
  Vector v0, v1;

  v0 = Vector(x0->x() - x1->x(), x0->y() - x1->y(), x0->z() - x1->z());
  v1 = Vector(x1->x() - x2->x(), x1->y() - x2->y(), x1->z() - x2->z());

  Vector n = cross(v0, v1);
  n.normalize();

  return (new Normal(n.x(), n.y(), n.z()));
}

const GLuint Model::numVertices() { return vertices.size(); }
const GLuint Model::numTriangles() { return triangles.size(); }

string Model::toString() {
  string s;
  char c[20];
/*
  for (set <Triangle *>::iterator tri = triangles.begin(); tri != triangles.end(); tri++, i++) {
    sprintf(c, "Triangle %d:\n", i);
    s += c;
    s += (*tri)->toString();
  }
*/
  for (VSI v = vertices.begin(); v != vertices.end(); v++) {
    sprintf(c, "Vertex %d:\n", (*v)->getID());
    s += c;
    s += (*v)->toString();
  }

  return s;
}


GLvoid Model::configureVectors(vector <Vertex *> * v, vector <GLint> * vi, vector <Color *> * c) {
  GLuint i;

  /*
    Creates colors for each vertex
  */
  if (c->size() != v->size()) {  // the color defaults to a light grey
    for (i = 0; i != v->size(); i++) {
      (*v)[i]->setColor(new Color(0.78f, 0.78f, 0.78f));
      vertices.insert((*v)[i]);
    }
  }
  else {
    for (i = 0; i != v->size(); i++) {
      (*v)[i]->setColor((*c)[i]);
      vertices.insert((*v)[i]);
    }
  }

  /*
    Creates triangles vector
  */
  for (i = 0; i != vi->size(); i+=3) {
    triangles.insert(new Triangle((*v)[(*vi)[i]],
                                  (*v)[(*vi)[i + 1]],
                                  (*v)[(*vi)[i + 2]]));
  }
}

VS * Model::getVertexList() { return &vertices; }
set <Triangle *> * Model::getTriangleList() { return &triangles; }

GLint Model::getNumCorners() { return numCorners; }
GLvoid Model::setNumCorners(GLint i) { numCorners = i; }
