// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef MODEL_H
#define MODEL_H

#include "options.h"

#include <cstdio>
#include <cmath>
#include <vector>
#include <map>
#include <string>
#include "Triangle.h"
#include "Vertex.h"
#include "Color.h"
#include "Normal.h"
#include <GL/gl.h>

//!  Model Class.  Holds information about a single model in a VRML Scene.
class Model {
  /*!
    The boolean value for vertex ordering. If it is true this Model has a
    clockwise vertex order, if false the order is counter/anti clockwise.
  */
  bool clockwise;

  //! The number of corners in this model. We never remove a corner.
  GLint numCorners;

  //! Helper function to compute the normals at each vertex.
  Normal * computeNormals(Vertex * x0, Vertex * x1, Vertex * x2);
  //! Helper function to set the information at each vertex.
  GLvoid configureVectors(std::vector <Vertex *> * v, std::vector <GLint> * vi, std::vector <Color *> * c);

  /*!
    This is what is read from the VRML file and does not need
    to be altered; it is a final data structure.
  */
  VS vertices;

  /*!
    This is the new vector that holds the altered information;
    this is a final data structure.
  */
  std::set <Triangle *> triangles;

 public:
  //! Constructor taking information for all vertices.
  Model(std::vector <Vertex *> * v, //!< List of vertices.
	std::vector <GLint> * vi, //!< List of vertex indices.
	std::vector <Color *> * c, //!< List of colors for each vertex.
	bool order);

  const GLuint numVertices(); //!< Returns the number of vertices.
  const GLuint numTriangles(); //!< Returns the number of triangles.

  std::string toString(); //!< Returns a string showing all information.

  VS * getVertexList(); //!< Returns a set of the verticies.
  std::set <Triangle *> * getTriangleList(); //<! Returns a set of the triangles.

  void addTriangle(Triangle * t);
  void removeTriangle(Triangle * t);

  void addVertex(Vertex * v);
  void removeVertex(Vertex * v);

  bool isClockwise();
  
  GLint getNumCorners();
  GLvoid setNumCorners(GLint i);
};

#endif
