// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Scene.h"
#define SCENE_DEBUG false

using namespace std;

Scene::Scene() {
  count = 0;
  backUp = 0.0f;
  center = Point();
  ROT_STEP = 0.0f;
  TRANS_STEP = 0.0f;
  FRUSTRUM_DEPTH = 0.0f;
  NORMAL_LENGTH = 0.0f;

  scene_is_null = true;
}


GLvoid Scene::addModel(Model * m) { models.insert(models.end(), *m); }
GLvoid Scene::removeModel(Model * m)
{ 
  vector<Model>::iterator mod;
  for( mod = models.begin(); mod != models.end() ; mod++)
    if(&(*mod) == m) break;
  models.erase(mod);
}

Model * Scene::m(GLint m) { return &models[m]; }
GLint Scene::numModels() { return models.size(); }

GLvoid Scene::addLight(Light * l) { lights.insert(lights.end(), *l); }
Light * Scene::l(GLint l) { return &lights[l]; }
GLint Scene::numLights() { return lights.size(); }

/*
  This function makes the appropriate changes to the way the viewport views the scene.
  Viewport's are dumb Scene's are smart.
  OR
  Viewport's know how to view.
  Scene's tell Viewport's what to view.
  Changes are made to variables that depend on the size of the Scene, such as:
  backUp
  center
  ROT_STEP
  TRANS_STEP
  FRUSTRUM_DEPTH
  NORMAL_LENGTH

  this funciton also detects corners in each model. we never want to remove a corner!
*/
GLvoid Scene::analyzeScene() {
  GLfloat totalX = 0.0f;
  GLfloat totalY = 0.0f;
  GLfloat totalZ = 0.0f;

  GLfloat startX = 0.0f;
  GLfloat startY = 0.0f;
  GLfloat startZ = 0.0f;

  GLfloat stopX = 0.0f;
  GLfloat stopY = 0.0f;
  GLfloat stopZ = 0.0f;

  for (vector <Model>::iterator model = models.begin();
       model != models.end();
       model++) {
    VS vertices = *(model->getVertexList());
    GLint numCorners = 0;

    for (VSI iter = vertices.begin(); iter != vertices.end(); iter++) {
      // get bounds of viewport
      GLfloat x, y, z;

      count++;
      x = (*iter)->x();
      y = (*iter)->y();
      z = (*iter)->z();

      // Useful for getting center of model: LOCAL variables
      totalX += x;
      totalY += y;
      totalZ += z;

      // find the greatest length of any side/plane
      if (startX > x)
	startX = x;
      else if (stopX < x)
	stopX = x;

      if (startY > y)
	startY = y;
      else if (stopY < y)
	stopY = y;

      if (startZ > z)
	startZ = z;
      else if (stopZ < z)
	stopZ = z;

      // determine corners
      // take the dot product of the two triangle normals that share the two studied vertices
      /*

                .v
               /|\
              / | \
             /  |a \
            /   |   \
           .----.----.
                n

        v is the vertice
        n is the neighbor
        a is the edge that shares v and n

        What we need is the dot product of the normal of the two triangles that share that edge.
        Then compare that number to see if it is an edge.
      */

      bool one_feature_edge = false;
      bool two_feature_edges = false;
      bool three_feature_edges = false;
      unsigned int nCount = 0;
      set <Vertex *> neighbors = *(*iter)->getNeighbors();
      for (set <Vertex *>::iterator neigh = neighbors.begin();
           neigh != neighbors.end();
           neigh++) {
        nCount++;
        set <Triangle *> * tris = (*iter)->getSharedTriangles(*neigh);
        set <Triangle *>::iterator t = tris->begin();

        // if there is only one triangle that they share it is a boundry (do not remove, for now)
        if (tris->size() == 1) {
          (*iter)->isCorner(true);
	  numCorners++;
          break;
        }

        if (tris->size() != 2) { continue; }

        Triangle * tri1 = (*t);
        t++;
        Triangle * tri2 = (*t);

        Normal * normal1 = tri1->getNormal();
        Normal * normal2 = tri2->getNormal();

        float result = dot(*normal1, *normal2);
        if (result < 1.55f) {
          if (one_feature_edge) { one_feature_edge = false; two_feature_edges = true; }
          else if (two_feature_edges) { two_feature_edges = false; three_feature_edges = true; }
          else if (three_feature_edges) {}    // more than three?
          else one_feature_edge = true;
        }

        if (three_feature_edges) {    // if we found three shared edges
          (*iter)->isCorner(true);
          numCorners++;
          break;
        }
        // if we found one feature edge and there are no more to find
        else if (one_feature_edge && neighbors.size() == nCount) {
          (*iter)->isCorner(true);
          numCorners++;
          break;
        }
      }
    }  // end of for each Vertex

    // make sure the model knows how many corners it contains
    model->setNumCorners(numCorners);
  }  // end of for each Model

  GLfloat sideX = fabs(stopX - startX);
  GLfloat sideY = fabs(stopY - startY);
  GLfloat sideZ = fabs(stopZ - startZ);

  if (sideX > sideY) {
    if (sideX > sideZ)
      backUp = sideX;
    else
      backUp = sideZ;
  }
  else if (sideY > sideZ)
    backUp = sideY;
  else
    backUp = sideZ;

  GLfloat x = totalX / count;
  GLfloat y = totalY / count;
  GLfloat z = totalZ / count;

  center = Point(x, y, z);

  ROT_STEP = backUp / 5;
  TRANS_STEP = backUp / 25;
  FRUSTRUM_DEPTH = backUp * 4;
  NORMAL_LENGTH = backUp / 25;

  /*
    This is for adjusting backUp so we can see all of the models.
    If we didn't do this we would not see the top's bottom's or side's of the model.
    This is only because the backUp is not large enough
  */
  backUp*=2.0f;

  // the scene is only fully functional after all the analysis has been completed
  scene_is_null = false;
}

bool Scene::isFullyFunctional() { return !scene_is_null; }

Point Scene::getCenter() { return center; }

GLfloat Scene::getBackUp() { return backUp; }

GLfloat Scene::getRotStep() { return ROT_STEP; }

GLfloat Scene::getTransStep() { return TRANS_STEP; }

GLfloat Scene::getFrustrumDepth() { return FRUSTRUM_DEPTH; }

GLfloat Scene::getNormalLength() { return NORMAL_LENGTH; }

GLint Scene::getCount() { return count; }

string Scene::toString() {
  string s;
  char c[100];
  GLint i = 0;

  s = "Models:\n";

  for(vector <Model>::iterator m = models.begin();
      m != models.end();
      m++, i++) {
    sprintf(c, "\nModel ID: %d\n", i);
    s += c;
    s += m->toString();
  }

  s += "\nLights:\n";
  i = 0;

  for(vector <Light>::iterator l = lights.begin();
      l != lights.end();
      l++, i++) {
    sprintf(c, "\nLight ID: %d\n", i);
    s += c;
    s += l->toString();
  }

  return s;
}
