// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef SCENE_H
#define SCENE_H

#include <iostream>
#include <vector>
#include <cmath>
#include "Model.h"
#include "Vertex.h"
#include "Point.h"
#include "Vector.h"
#include "Light.h"

//! Class definition for a Scene
class Scene {

  //! List of Models in the Scene
  std::vector <Model> models;

  //! List of Lights in the Scene
  std::vector <Light> lights;

  GLfloat ROT_STEP;
  GLfloat TRANS_STEP;
  GLfloat FRUSTRUM_DEPTH;
  GLfloat NORMAL_LENGTH;
  Point center; //!< Center of the Scene (subtracted from all coordinates)
  GLfloat backUp;
  GLint count;
  GLfloat diffuseRed;
  GLfloat diffuseGreen;
  GLfloat diffuseBlue;
  bool scene_is_null;

 public:
  Scene(); //!< Constructor

  //! Add a Model to the Scene
  GLvoid addModel(Model * m);

  //! Remove a Model in the Scene. This occurs when we removed all the vertices during the decimation process.
  GLvoid removeModel(Model * m);

  //! Get the mth Model in the Scene
  Model * m(GLint m);

  //! Get the number of Models in the Scene
  GLint numModels();

  //! Add a Light to the Scene
  GLvoid addLight(Light * l);

  //! Get the lth Light in the Scene
  Light * l(GLint l);

  //! Get the number of Lights in the Scene
  GLint numLights();

  //! Get the center Point of the Scene
  Point getCenter();

  //! Get the distace to back up the camera for display (as calculated in analyzeScene)
  GLfloat getBackUp();

  //! Get the angle to rotate through for display (as calculated in analyzeScene)
  GLfloat getRotStep();

  //! Get the distance to translate through for display (as calculated in analyzeScene)
  GLfloat getTransStep();

  //! Get the depth of the frustum for display (as calculated in analyzeScene)
  GLfloat getFrustrumDepth();

  //! Get the length of the Normals for display (as calculated in analyzeScene)
  GLfloat getNormalLength();

  //! Get the number of Models in the Scene
  GLint getCount();

  //! Center the Scene in the Viewport and set the Normal lengths for display
  GLvoid analyzeScene();

  //! Are we ready to render?
  bool isFullyFunctional();

  //! Get information for the Scene in string form
  std::string toString();
};

#endif
