// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Triangle.h"

using namespace std;

Triangle::Triangle(Vertex * v0, Vertex * v1, Vertex * v2) {
  replaced_before = false;

  calculateColor(v0, v1, v2);
  calculateNormal(v0, v1, v2);
  calculateCenter(v0, v1, v2);

  v[0] = v0;
  v[1] = v1;
  v[2] = v2;

  v[0]->addTriangle(this);
  v[1]->addTriangle(this);
  v[2]->addTriangle(this);
}

bool Triangle::replaced() { return replaced_before; }

bool Triangle::replace(Vertex * old_vertex, Vertex * new_vertex) {
  replaced_before = false;

  // a triangle structure does not exist any longer when two of the three vertices are the same
  if (contains(new_vertex) && contains(old_vertex)) {
    return false;
  }

  if (v[0] == old_vertex && v[1] == old_vertex) {
    v[0]->removeTriangle(this);
    return false;
  }
  if (v[1] == old_vertex && v[2] == old_vertex) {
    v[1]->removeTriangle(this);
    return false;
  }
  if (v[0] == old_vertex && v[2] == old_vertex) {
    v[2]->removeTriangle(this);
    return false;
  }

  if (v[0] == old_vertex) {
    v[0] = new_vertex;
    v[0]->addTriangle(this);
  }
  else if (v[1] == old_vertex) {
    v[1] = new_vertex;
    v[1]->addTriangle(this);
  }
  else if (v[2] == old_vertex) {
    v[2] = new_vertex;
    v[2]->addTriangle(this);
  }
  else return false;

  v[0]->replace(old_vertex, new_vertex);
  v[1]->replace(old_vertex, new_vertex);
  v[2]->replace(old_vertex, new_vertex);

  calculateColor(v[0], v[1], v[2]);
  calculateNormal(v[0], v[1], v[2]);
  calculateCenter(v[0], v[1], v[2]);

  replaced_before = true;
  return true;
}

bool Triangle::contains(Vertex * v1) {
  if (v[0] == v1) return true;
  if (v[1] == v1) return true;
  if (v[2] == v1) return true;
  return false;
}

/*
  Get the average color of all the points
*/
GLvoid Triangle::calculateColor(Vertex * v0, Vertex * v1, Vertex * v2) {
  color = Color((v0->getColor()->r() + v1->getColor()->r() + v2->getColor()->r()) / 3,
		(v0->getColor()->g() + v1->getColor()->g() + v2->getColor()->g()) / 3,
		(v0->getColor()->b() + v1->getColor()->b() + v2->getColor()->b()) / 3,
		(v0->getColor()->a() + v1->getColor()->a() + v2->getColor()->a()) / 3);
}

/*
  Get the average normal of all the points
*/
GLvoid Triangle::calculateNormal(Vertex * v0, Vertex * v1, Vertex * v2) {
  Vector x0, x1;

  x0 = Vector(v0->x() - v1->x(), v0->y() - v1->y(), v0->z() - v1->z());
  x1 = Vector(v1->x() - v2->x(), v1->y() - v2->y(), v1->z() - v2->z());

  x0 = cross(x0, x1);
  x0.normalize();

  normal.x() = x0.x();
  normal.y() = x0.y();
  normal.z() = x0.z();
}

/*
  Calculates a point in 3D space, which is the center of the Triangle
*/
GLvoid Triangle::calculateCenter(Vertex * v0, Vertex * v1, Vertex * v2) {
  center = Point((v0->x() + v1->x() + v2->x()) / 3.0f,
		 (v0->y() + v1->y() + v2->y()) / 3.0f,
		 (v0->z() + v1->z() + v2->z()) / 3.0f);
}

Vertex * Triangle::v0() { return v[0]; }
Vertex * Triangle::v1() { return v[1]; }
Vertex * Triangle::v2() { return v[2]; }

GLvoid Triangle::setColor(Color * c) {
  color = *c;
  v[0]->setColor(c);
  v[1]->setColor(c);
  v[2]->setColor(c);
}
Color * Triangle::getColor() { return &color; }

GLvoid Triangle::setNormal(Normal * n) {
  normal = *n;
}
Normal * Triangle::getNormal() { return &normal; }

GLvoid Triangle::setCenter(Point * c) { center = *c; }
Point * Triangle::getCenter() { return &center; }

string Triangle::toString() {
  string s;

  s = "Vertex 0:\n" + v[0]->toString() + "\n";
  s += "Vertex 1:\n" + v[1]->toString() + "\n";
  s += "Vertex 2:\n" + v[2]->toString() + "\n";

  return s;
}
