// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// seperate scene from vrml parser
// Need to give up and return something after a timeout period has arrived
// The parser must initiate the exit call

/*
  Normals are read like this:
  normalIndex = coordIndex, because they both have the same amount of triangles
  except the indexes point to different lists
  normalIndex -> normals
  coordIndex -> vertices
*/

#include "VRMLParser.h"

#define VRML_PARSER_DEBUG

using namespace std;




VRMLParser::VRMLParser() {
  translate = Vector(0.0f, 0.0f, 0.0f);
  numVertex = 0;
  clockwise = false;
}




bool VRMLParser::write(Scene * scene, string filename) {

  // Open file
#ifdef VRML_PARSER_DEBUG
    cout << "Opening file " << filename << ": ";
    cout.flush();
#endif //VRML_PARSER_DEBUG

  out.open(filename.c_str());
  if (out.bad()) return false;

  // Header
  out << "#VRML V2.0 utf8\n\n# Produced by Mesh Decimation 1.0\n\n";

  // For each Model
  for(GLint model = 0 ; model < scene->numModels() ; model++) {

    VS * vertices = scene->m(model)->getVertexList();
    VSI vertex;
    set<Triangle *> * triangles = scene->m(model)->getTriangleList();
    set<Triangle *>::iterator triangle;

    // Model top matter
    out << "DEF Model" << model << " Transform {\nchildren [\nShape {\n";
    out << "  geometry IndexedFaceSet {\n";
	out << "  ccw " << (scene->m(model)->isClockwise()?"FALSE":"TRUE")<< "\n  solid TRUE\n";
    // Color
    out << "  colorPerVertex TRUE\n  color Color {\n    color [\n";
    for(vertex = vertices->begin();
        vertex != vertices->end();
        vertex++)
      out << "      " << (*vertex)->getColor()->r() << " "
          << (*vertex)->getColor()->g() << " "
          << (*vertex)->getColor()->b() << ",\n";
    out << "    ]\n  }\n";

    map <Vertex *, int> vertex_index;

    // Points
    out << "  coord Coordinate {\n";
    out << "    point [\n";
    // this is a special for loop
    // in order to make the coordIndex below, we need to create index numbers
    // that point to the vertices. we use 'int i' to do this
    // this is needed since getID returns and ID that is NOT the index number
    int i = 0;
    for(vertex = vertices->begin(); vertex != vertices->end(); vertex++, i++) {
      out << "      " << (*vertex)->x() << " "
          << (*vertex)->y() << " "
          << (*vertex)->z() << ",\n";
      vertex_index[*vertex] = i;
    }
    out << "    ]\n  }\n";

    // Triangles
    // now we need to retrieve the index numbers from the map
    //out << "  coord Coordinate {\n";
    out << "  coordIndex [\n";
    for(triangle = triangles->begin(); triangle != triangles->end(); triangle++) {
      out << "    " << vertex_index[(*triangle)->v0()] << ", "
          << vertex_index[(*triangle)->v1()] << ", "
          << vertex_index[(*triangle)->v2()] << ", -1,\n";
		}
    out << "  ]\n  }\n";

    out << "\n}\n]\n}\n";
  }

  // Close the file
  out.flush();
  out.close();

  return true;
}




Scene * VRMLParser::parse(string filename) {

  Scene * s = new Scene();

  int type = 0;
  const int vrml1 = 1;
  const int vrml2 = 2;

  // open file
#ifdef VRML_PARSER_DEBUG
    cout << "Opening file " << filename << ": ";
    cout.flush();
#endif //VRML_PARSER_DEBUG

  in.open(filename.c_str());
  if (in.bad()) return s;
  
#ifdef VRML_PARSER_DEBUG
    cout << "done\nCreating temporary data structures: ";
    cout.flush();
#endif //VRML_PARSER_DEBUG

  verticesFinished = false;
  coordIndexFinished = false;
  colorsFinished = false;
  diffuseColorFinished = false;
  foundTranslation = false;

  // used to get lines from the file
  char CurLine[ 1024 ];
  char *c;

  // check the header and make sure the file is a VRML file 
#ifdef VRML_PARSER_DEBUG
    cout << "done\nChecking file header: ";
    cout.flush();
#endif //VRML_PARSER_DEBUG

  in.getline( CurLine, 1024, '\n' );
  if ( !strncmp( CurLine, "#VRML V1.0 ascii", 16 ) ) type = vrml1;
  else if ( !strncmp( CurLine, "#VRML V2.0 utf8", 15 ) ) type = vrml2;
  else if ( !strncmp( CurLine, "#Inventor V2.0 ascii", 20 ) ) { type = vrml2; }
  else if ( !strncmp( CurLine, "#Inventor V2.1 ascii", 20 ) ) { type = vrml2; }
  else return s;
#ifdef VRML_PARSER_DEBUG
  cout << "VRML " << type << ".0" << endl;
#endif //VRML_PARSER_DEBUG

  // read the data
  while(!in.eof()) {

    // compact the line - strip out worthless stuff 
    in >> CurLine;
    for(c = CurLine; isspace( *c ); c++);

    // if the line is a comment ignore the rest of the line 
    if (*c == '#') {
      in.getline(CurLine, 1024, '\n');
      continue;
    }

    // read the token
    if (!strcmp(CurLine, "vertexOrdering")) {
#ifdef VRML_PARSER_DEBUG
	cout << "Found vertex ordering: ";
	cout.flush();
#endif //VRML_PARSER_DEBUG

      in >> CurLine;
      if (!strcmp(CurLine, "CLOCKWISE")) {
#ifdef VRML_PARSER_DEBUG
  cout << "clockwise";
#endif //VRML_PARSER_DEBUG
	clockwise = true;
      }
      else if (!strcmp(CurLine, "COUNTERCLOCKWISE")) {
#ifdef VRML_PARSER_DEBUG
  cout << "counter clockwise";
#endif //VRML_PARSER_DEBUG
	clockwise = false;
      }
      else {
#ifdef VRML_PARSER_DEBUG
  cout << "unknown ordering.  Ignoring";
#endif //VRML_PARSER_DEBUG
      }
#ifdef VRML_PARSER_DEBUG
  cout << "." << endl;
#endif //VRML_PARSER_DEBUG
    }
    else if (!strcmp(CurLine, "ccw")) {
#ifdef VRML_PARSER_DEBUG
	cout << "Found vertex ordering: ";
	cout.flush();
#endif //VRML_PARSER_DEBUG

      in >> CurLine;
      if (!strcmp(CurLine, "FALSE")) {
#ifdef VRML_PARSER_DEBUG
  cout << "clockwise";
#endif //VRML_PARSER_DEBUG
	clockwise = true;
      }
      else if (!strcmp(CurLine, "TRUE")) {
#ifdef VRML_PARSER_DEBUG
  cout << "counter clockwise";
#endif //VRML_PARSER_DEBUG
	clockwise = false;
      }
      else {
#ifdef VRML_PARSER_DEBUG
  cout << "unknown ordering.  Ignoring";
#endif //VRML_PARSER_DEBUG
      }
#ifdef VRML_PARSER_DEBUG
cout << "." << endl;
#endif //VRML_PARSER_DEBUG
    }
    else if (!strcmp(CurLine, "translation")) {
      if (foundTranslation)
	addModelToScene(s);

      foundTranslation = true;

#ifdef VRML_PARSER_DEBUG
      cout << "Found translation value"; cout.flush();
#endif //VRML_PARSER_DEBUG

      GLfloat x, y, z;
      in >> x >> y >> z;
      translate = Vector(x, y, z);
#ifdef VRML_PARSER_DEBUG
cout << "." << endl;
#endif //VRML_PARSER_DEBUG
    }
    else if (!strcmp(CurLine, "diffuseColor")) {
      // Test to see if the parser has been here before      
      if (diffuseColorFinished)
	addModelToScene(s);
      // This stays true because we are in here either way
      diffuseColorFinished = true;

#ifdef VRML_PARSER_DEBUG
      cout << "Found diffuse color"; cout.flush();
#endif //VRML_PARSER_DEBUG

      GLfloat x, y, z;
      char CurChar;

      in >> CurChar;
      if( CurChar != '[' ) {
	in.putback(CurChar);
	in >> x >> y >> z;
      }
      else {
	x = 0.7f;
	y = 0.7f;
	z = 0.7f;
      }

      diffuseColor = new Color(x, y, z);
#ifdef VRML_PARSER_DEBUG
  cout << "." << endl;
#endif //VRML_PARSER_DEBUG
    }
    else if (!strcmp(CurLine, "color" )) {
      // Test to see if the parser has been here before
      if (colorsFinished)
	addModelToScene(s);

      if (!diffuseColorFinished) {
	// This stays true because we are in here either way
	colorsFinished = true;

	// No matter what happens above the parser continues to get the vertices
	colors = parseColor();
      }
    }
    else if (!strcmp(CurLine, "point")) {
      // Test to see if the parser has been here before
      if (verticesFinished)
	addModelToScene(s);

      // This stays true because we are in here either way
      verticesFinished = true;

      // No matter what happens above the parser continues to get the vertices
      vertices = parseVertices();
    }
    else if (!strcmp(CurLine, "coordIndex")) {
      // Test to see if the parser has been here before
      if (coordIndexFinished)
	addModelToScene(s);

      // In VRML the vertices are defined before the coordIndex
      if (verticesFinished) {
	coordIndex = parseCoordIndex();
	// This stays true because we are in here either way
	coordIndexFinished = true;
      }
      else {
	cout << "This file is bad. coordIndex found before points: exiting." << endl;
	exit(0);
      }
    }
  }

  // Add the last model
  addModelToScene(s);

  // This is where we find the center of the Scene and where we deterimine
  // the distance from the center of the Scene for easier viewing.
#ifdef VRML_PARSER_DEBUG
    cout << "Analysing Scene: ";
    cout.flush();
#endif //VRML_PARSER_DEBUG
  s->analyzeScene();
#ifdef VRML_PARSER_DEBUG
  cout << "done." << endl;
#endif //VRML_PARSER_DEBUG

  // close the file
#ifdef VRML_PARSER_DEBUG
    cout << "Closing file: ";
    cout.flush();
#endif //VRML_PARSER_DEBUG
  in.close();
#ifdef VRML_PARSER_DEBUG
  cout << "done." << endl;
#endif //VRML_PARSER_DEBUG

  return s;
}




vector <Vertex *> VRMLParser::parseVertices() {
  GLfloat x, y, z;
  char CurChar = ' ';
  vector <Vertex *> v;
  Vertex * vert;

#ifdef VRML_PARSER_DEBUG
  cout << "Found vertices: "; cout.flush();
#endif //VRML_PARSER_DEBUG

  while (CurChar != '[') in >> CurChar;

  while (true) {
    // read the relevant info
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> x;
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> y;
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> z;

    vert = new Vertex(translate.x() + x, translate.y() + y, translate.z() + z);
    vert->setID(numVertex);
    numVertex++;
    v.insert(v.end(), vert);

    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> CurChar;
    if( CurChar != ']' ) in.putback(CurChar);
    else break;
  }

  translate = Vector(0.0f, 0.0f, 0.0f);
#ifdef VRML_PARSER_DEBUG
    cout << "parsed " << v.size() << " vertices." << endl;
#endif //VRML_PARSER_DEBUG

  return v;
}




vector <int> VRMLParser::parseCoordIndex() {
  GLint i, j, k, l;
  char CurChar = ' ';
  vector <int> ci;

#ifdef VRML_PARSER_DEBUG
  cout << "Found vertex index: "; cout.flush();
#endif //VRML_PARSER_DEBUG

  while (CurChar != '[') in >> CurChar;

  if (clockwise) {
    while (true) {
      // read the relevant info 
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> i;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> j;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> k;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> l;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      // add the triangle
      ci.insert(ci.end(), k);
      ci.insert(ci.end(), j);
      ci.insert(ci.end(), i);
      in >> CurChar;
      if( CurChar != ']' ) in.putback(CurChar);
      else break;
    }
  }
  else {
    while (true) {
      // read the relevant info 
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> i;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> j;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> k;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      in >> l;
      in >> CurChar;
      if( CurChar != ',' ) in.putback(CurChar);
      // add the triangle
      ci.insert(ci.end(), i);
      ci.insert(ci.end(), j);
      ci.insert(ci.end(), k);
      in >> CurChar;
      if( CurChar != ']' ) in.putback(CurChar);
      else break;
    }
  }

#ifdef VRML_PARSER_DEBUG
    cout << "parsed " << (ci.size() / 3)
	 << " triangles in the vertex index." << endl;
#endif //VRML_PARSER_DEBUG

  return ci;
}




vector <Color *> VRMLParser::parseColor() {
  GLfloat x, y, z;
  char CurChar = ' ';
  vector <Color *> c;

#ifdef VRML_PARSER_DEBUG
  cout << "Found colors: "; cout.flush();
#endif //VRML_PARSER_DEBUG

  while (CurChar != '[') in >> CurChar;

  while (true) {
    // read the relevant info 
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> x;
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> y;
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> z;
    in >> CurChar;
    if( CurChar != ',' ) in.putback(CurChar);
    in >> CurChar;
    // add the color to the color list
    c.insert(c.end(), new Color(x, y, z));
    if( CurChar != ']' ) in.putback(CurChar);
    else break;
  }

#ifdef VRML_PARSER_DEBUG
    cout << "parsed " << c.size() << " colors." << endl;
#endif //VRML_PARSER_DEBUG

  return c;
}




/*
  This function makes all of the colors the same.
  It creates as many colors as vertices,
  but vertices has to be created first.
*/
vector <Color *> VRMLParser::setColor(Color * c) {
  vector <Color *> cList;

  for (GLuint i = 0; i != vertices.size(); i++)
    // add the color to the color list
    cList.insert(cList.end(), c);

  return cList;
}




GLvoid VRMLParser::addModelToScene(Scene * s) {
#ifdef VRML_PARSER_DEBUG
    cout << "Adding Model to the Scene: ";
    cout.flush();
#endif //VRML_PARSER_DEBUG

  if (diffuseColorFinished) {
    colors = setColor(diffuseColor);
    diffuseColor = new Color();
  }

  Model m = Model(&vertices, &coordIndex, &colors, clockwise);

  // Add the model to the Scene
  s->addModel(&m);

#ifdef VRML_PARSER_DEBUG
  cout << "done." << endl;
#endif //VRML_PARSER_DEBUG

  verticesFinished = false;
  coordIndexFinished = false;
  colorsFinished = false;
  diffuseColorFinished = false;
  foundTranslation = false;
}
