// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "Vertex.h"

using namespace std;

Vertex::Vertex(GLfloat x, GLfloat y, GLfloat z, GLfloat w) : Point(x, y, z, w) {
  color = Color(0.5f, 0.5f, 0.5f, 1.0f);
  normal = Normal();
  corner = false;
  ID = -1;
  distance = 0.0f;
}

GLvoid Vertex::setID(GLint id) {
  ID = id;
}
GLint Vertex::getID() {
  return ID; }

// addNeighbor and removeNeighbor must balance in terms of modifications to the mesh
GLvoid Vertex::addNeighbor(Vertex * v) {
  if (v != this) {
    neighbors.insert(v);
    if (!v->contains(this)) v->addNeighbor(this);

    calculateNormal();
    calculateDistance();    // if it succeeded we must recalculate the distance
  }
}

bool Vertex::removeNeighbor(Vertex * v) {
  if (contains(v)) {
    neighbors.erase(v);

    if (v->contains(this)) v->removeNeighbor(this);

    // we no longer need the triangles that are associated with
    // that neighbor, so we remove them
    set <Triangle *> * tris = getSharedTriangles(v);
    for (set <Triangle *>::iterator t = tris->begin();
         t != tris->end();
         t++) {
      removeTriangle(*t);
    }
  }
  else return false;

  if( neighbors.size() == 0 ) return false;

  calculateNormal();
  calculateDistance();    // if it succeeded we must recalculate the distance
  return true;
}

bool Vertex::replace(Vertex * old_vertex, Vertex * new_vertex) {
  if(removeNeighbor(old_vertex)) addNeighbor(new_vertex);
  else return false;
  return true;
}

/*
  We need to keep this data structure complete.
  So if we are asked to add a Triangle we need to know if
  it is our triangle. If it is then we need to make sure the
  other vertices in the Triangle know we are neighbors.
*/
GLvoid Vertex::addTriangle(Triangle * t) {
  if (t->contains(this)) {
    triangles.insert(t);

    Vertex * v0 = t->v0();
    Vertex * v1 = t->v1();
    Vertex * v2 = t->v2();

    // tell the other vertices in the Triangle
    // that this is your new neighbor
    v0->addNeighbor(v1);
    v0->addNeighbor(v2);

    v1->addNeighbor(v0);
    v1->addNeighbor(v2);

    v2->addNeighbor(v0);
    v2->addNeighbor(v1);

    if (!v0->contains(t)) v0->addTriangle(t);
    if (!v1->contains(t)) v1->addTriangle(t);
    if (!v2->contains(t)) v2->addTriangle(t);
  }
}

// after erasing the Triangle from the set. we can find any Vertex that is no longer a neighbor by
// getSharedTriangles(Vertex) if it returns none, then it is no longer a neighbor
// but be carefull not to access a vertex that has been deleted
bool Vertex::removeTriangle(Triangle * t) {
  set <Triangle *>::iterator i = triangles.find(t);
  if (i != triangles.end()) {
    triangles.erase(i);
  }
  else return false;

  return true;
}


bool Vertex::isCorner() {
  return corner; }
GLvoid Vertex::isCorner(bool b) {
  corner = b; }

/*
  Helper function for decimate. This function calculates the
  distance of this Vertex to the average plane defined by neighbors.
*/
GLvoid Vertex::calculateDistance() {

  // Create the point in question Vector
  Vector vertex = Vector(x(), y(), z());

  /*
    get average of the neighbors: totalX, totalY, totalZ
    get average of the neighbors normals: totalX, totalY, totalZ
  */
  GLfloat totalX = 0.0f;
  GLfloat totalY = 0.0f;
  GLfloat totalZ = 0.0f;

  GLfloat normTotalX = 0.0f;
  GLfloat normTotalY = 0.0f;
  GLfloat normTotalZ = 0.0f;

  for (set <Vertex *>::iterator iter = neighbors.begin(); iter != neighbors.end(); iter++) {
    totalX += (*iter)->x();
    totalY += (*iter)->y();
    totalZ += (*iter)->z();

    normTotalX += (*iter)->getNormal()->x();
    normTotalY += (*iter)->getNormal()->y();
    normTotalZ += (*iter)->getNormal()->z();
  }

  // Create the average point Vector
  Vector avgP = Vector(totalX/neighbors.size(), totalY/neighbors.size(), totalZ/neighbors.size());

  // Create the average normal Vector
  Vector avgN = Vector(normTotalX/neighbors.size(), normTotalY/neighbors.size(), normTotalZ/neighbors.size());
  avgN.normalize();
  distance = fabs(dot(vertex, avgN) - dot(avgP, avgN));
}

GLfloat Vertex::getDistance() {
  return distance; }

GLvoid Vertex::setColor(Color * c) {
  color = *c; }
GLvoid Vertex::setNormal(Normal * n) {
  normal = *n; }

/*!
  There are two ways a vertex can get it's normal
  1. It can be given it with the setNormal() operation above
  2. It can calculate it if the neighbors have been specified (below)
*/
GLvoid Vertex::calculateNormal() {
  if (triangles.size() == 0) {
    normal = Normal(0.0f, 0.0f, 0.0f);
  }
  else {
    GLfloat x = 0.0f, y = 0.0f, z = 0.0f;

    for (set <Triangle *>::iterator tri = triangles.begin();
	 tri != triangles.end();
	 tri++) {
      x += (*tri)->getNormal()->x();
      y += (*tri)->getNormal()->y();
      z += (*tri)->getNormal()->z();
    }

    normal = Normal(x, y, z);
    normal.normalize();
  }
}

Color * Vertex::getColor() {
  return &color; }
Normal * Vertex::getNormal() {
  return &normal; }

set <Triangle *> * Vertex::getTriangles() {
  return &triangles; }
set <Vertex *> * Vertex::getNeighbors() {
  return &neighbors; }

bool Vertex::contains(Vertex * v) {
  if (neighbors.find(v) != neighbors.end()) return true;
  else return false;
}

bool Vertex::contains(Triangle * t) {
  if (triangles.find(t) != triangles.end()) return true;
  else return false;
}

string Vertex::toString() {
  string s;

  s = "Point:\t" + Point::toString() + "\n";
  s += "Normal:\t" + normal.toString() + "\n";
  s += "Color:\t" + color.toString() + "\n";

  return s;
}

set <Triangle *> * Vertex::getSharedTriangles(Vertex * v) {
  shared.clear();
  for (set <Triangle *>::iterator t1 = triangles.begin();
       t1 != triangles.end();
       t1++) {
    if ((*t1)->contains(v)) {
      // this means that t is one of the triangles we need
      shared.insert(*t1);
    }
  }

  set <Triangle *> * triList = v->getTriangles();
  for (set <Triangle *>::iterator t2 = triList->begin();
       t2 != triList->end();
       t2++) {
    if ((*t2)->contains(this)) {
      // this means that t is one of the triangles we need
      shared.insert(*t2);
    }
  }

  return &shared;
}
