// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef VERTEX_H
#define VERTEX_H

class Vertex;

#include "options.h"
#include <iostream>
#include <vector>
#include <set>
#include <cmath>
#include "Triangle.h"
#include "Color.h"
#include "Normal.h"
#include "Point.h"
#include <GL/gl.h>

//! Holds information about a Vertex
class Vertex : public Point {
  //! List of triangles of which this Vertex is a part
  std::set <Triangle *> triangles;

  //! List of Vertices that are also members of those triangles
  std::set <Vertex *> neighbors;

  //! List of shared Triangles, which is returned in getSharedTriangles()
  std::set <Triangle *> shared;

  //! The distance of this Vertex from the average plane made by its neighbors
  GLfloat distance;

  //! The Color of this Vertex
  Color color;

  //! The Normal at this Vertex
  Normal normal;

  //! Determines if this vertex is a corner. We do not remove corners.
  bool corner;

  //! The ID of this Vertex
  GLint ID;

  //! Helper function to calculate the Normal for this Vertex
  GLvoid calculateNormal();

  //! Helper function to calculate the distance of this Vertex
  GLvoid calculateDistance();

 public:

  //! Constructor
  Vertex(
    float x, //!< The x coordinate of the Vertex
    float y, //!< The y coordinate of the Vertex
    float z, //!< The z coordinate of the Vertex
    float w = 1.0 //!< The w coordinate (space) of the Vertex (should remain 1.0 - need to normalize if not)
  );

  //! Get the distance of the Vertex to its neighbors' average plane
  GLfloat getDistance();

  //! Returns corner boolean
  bool isCorner();
  
  //! Sets corner boolean
  GLvoid isCorner(bool b);

  //! Add a Vertex to the list of neighboring Verticies
  GLvoid addNeighbor(Vertex * v);

  //! Add a Triangle to the list of member Triangles
  GLvoid addTriangle(Triangle * t);

  //! Remove a Vertex from the list of neighboring Vertices
  bool removeNeighbor(Vertex * v);

  //! Remove a Triangle from the list of member Triangles
  bool removeTriangle(Triangle * t);

  //! Set the Color of this Vertex
  GLvoid setColor(Color * c);

  //! Set the Normal of this Vertex
  GLvoid setNormal(Normal * n);

  //! Replace the Vertex v1 in the list of neigbors with Vertex v2
  bool replace(Vertex * v1, Vertex * v2);

  //! See if Vertex v is a member of the list of neighbors for this Vertex
  bool contains(Vertex * v);

  //! See if Triangle t is a member of the list of triangles for this Vertex
  bool contains(Triangle * t);

  //! Get the Color of the Vertex
  Color * getColor();

  //! Get the Normal of the Vertex
  Normal * getNormal();

  //! Get the list of Triangles this Vertex is a member of
  std::set <Triangle *> * getTriangles();

  //! Get the list of Vertices with are neighbors of this Vertex
  std::set <Vertex *> * getNeighbors();

  //! Show key information about this Vertex in string form
  std::string toString();

  //! Get the ID of this Vertex
  GLint getID();

  //! Set the ID of this Vertex
  GLvoid setID(GLint id);
  
  //! Returns the Triangles (always two) shared by this Vertex and v (passed)
  std::set <Triangle *> * getSharedTriangles(Vertex * v);
};

struct less_than_distance {
  bool operator () (Vertex * v0, Vertex * v1) const {
    return (v0->getDistance() < v1->getDistance());
  }
};

typedef std::set <Vertex *> VS;
typedef std::set <Vertex *>::iterator VSI;

#endif
