// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/*
  Should display information in a HUD type format, and have a switch to turn it off

  used display lists - 11/12/2000
  used vertex arrays - 11/12/2000
*/
#include "Viewport.h"

// Whether or not to display debugging messages
#define VIEWPORT_DEBUG false

// Use the standard c++ libs
using namespace std;

void Viewport::startingView() {
  /*
    Default transformation and rotation settings
  */
  normalLength = scene->getNormalLength();
  frustrumDepth = scene->getFrustrumDepth();
  center = scene->getCenter();
  backUp = scene->getBackUp();
  TRANS_STEP = scene->getTransStep();
  ROT_STEP = scene->getRotStep();

  X_Rot = 0.0f;
  Y_Rot = 0.0f;
  Z_Rot = 0.0f;

  // We must go the opposite direction from the center
  X_Off = -center.x();
  Y_Off = -center.y();
  Z_Off = -center.z() - backUp;
}

void Viewport::defaultInit() {
  /*
    Default viewport drawing settings
  */
  lighting = true;
  smooth = false;
  draw_points = false;
  draw_edges = false;
  draw_triangles = true;
  normals = false;
  vert_normals = false;

  /*
    Default OpenGL initialization
  */
  glLineWidth(2.0f);
  glPointSize(3.0f);
  glEnable(GL_CULL_FACE);

#ifdef USE_DISPLAY_LISTS
  /*
    Set up display lists
  */
  partList = glGenLists(1);
  MODEL_LIST = partList;

  glNewList(MODEL_LIST, GL_COMPILE);
    drawModel();
  glEndList();
#endif
}


Viewport::Viewport(string file, GLint id, bool GUI) {
  loaded = load(file);

  if (GUI) {
    startx = 0;
    starty = 0;
    width = DEFAULT_WIDTH;
    height = DEFAULT_HEIGHT;
    viewportID = id;
    control = MODEL_CONTROL;

    defaultInit();
    startingView();
  }
}


Viewport::~Viewport() {
#ifdef USE_DISPLAY_LISTS
  glDeleteLists(partList, 1);
#endif
}


bool Viewport::save(string file) {
  if(file == "NONE_SPECIFIED")
    file = filename; // Default to file we loaded the model from
  else
    filename = file; // Set the default to the new filename
  
  VRMLParser parser;

  if( scene->isFullyFunctional() )
    return parser.write(scene, file);
  else
    return false;
}


bool Viewport::load(string file) {
  VRMLParser parser;

  scene = parser.parse(file);
  filename = file;
  return scene->isFullyFunctional();
}


bool Viewport::reload() {
  return load(filename);
}


bool Viewport::resetSize(GLint x, GLint y, GLint w, GLint h) {
  startx = x;
  starty = y;
  width = w;
  height = h;

  return true;
}


/*!
  This function returns the most how many vertices we are allowed to remove.
  The variable 'd' determines how much to remove. If the most vertices to
  remove is greater than the number of vertices in all the models minus a
  padding of MIN_SCENE_SIZE (the least number of vertices we can decimate
  to, i.e. the smallest the scene will ever be), then we return the total
  number of vertices in the model - MIN_SCENE_SIZE.
*/
GLint Viewport::getMax(GLfloat d, GLint TYPE) {
  GLint num = 0;
  GLint total_num_corners = 0;
  GLint i;

  // get the measure that was specified
  switch (TYPE) {
  case PERCENTAGE:    // if d is a percentage calculate maxNumRemoved
    for (i = 0; i != scene->numModels(); i++) {
      total_num_corners += scene->m(i)->getNumCorners();
      num += scene->m(i)->getVertexList()->size();
    }

    if ((GLint)(num * d) >= (num - total_num_corners)) return (num - total_num_corners);
    else return (GLint)(num * d);

  case NUMBER:
    for (i = 0; i != scene->numModels(); i++) {
      GLint numCorners = scene->m(i)->getNumCorners();
      num += (scene->m(i)->getVertexList()->size() - numCorners);
    }

    if (d >= num) return num;
    return (GLint)d;

  default:
    cout << "Viewport: ERROR - TYPE not specified to decimate(GLfloat d, GLint TYPE)" << endl;
    exit(0);
    break;    // for integrity
  }
}

/*
  Possible TYPE: (as defined in Viewport.h)
  PERCENTAGE: the value 'd' is a percentage of vertices to be removed.
  NUMBER: the value 'd' is the requested number of vertices to remove.
*/
bool Viewport::decimate(GLfloat d, GLint TYPE) {
  Decimate dec = Decimate();
  return dec.remove(scene, getMax(d, TYPE));
}


/*
  Draws the normals for the model on each triangle
*/
GLvoid Viewport::drawNormals() {
  set <Triangle *> triangles;
  set <Triangle *>::iterator tri;
  Normal * norm;
  Point * c;

  for (GLint i = 0; i != scene->numModels(); i++) {
    triangles = *(scene->m(i)->getTriangleList());

    for (tri = triangles.begin(); tri != triangles.end(); tri++) {
      norm = (*tri)->getNormal();
      c = (*tri)->getCenter();
      glVertex3f(c->x(), c->y(), c->z());
      glVertex3f((normalLength * norm->x()) + c->x(),
		 (normalLength * norm->y()) + c->y(),
		 (normalLength * norm->z()) + c->z());
    }
  }
}


/*
  Draws the normals for the model on each vertice
*/
GLvoid Viewport::drawVertNormals() {
  VS vertices;
  VSI v;

  for (GLint i = 0; i != scene->numModels(); i++) {
    vertices = *(scene->m(i)->getVertexList());

    for (v = vertices.begin(); v != vertices.end(); v++) {
      glVertex3f((*v)->x(), (*v)->y(), (*v)->z());
      glVertex3f((normalLength * ((*v)->getNormal()->x())) + (*v)->x(),
                 (normalLength * ((*v)->getNormal()->y())) + (*v)->y(),
                 (normalLength * ((*v)->getNormal()->z())) + (*v)->z());
    }
  }
}

/*
  The O(n*n) 'big-oh-n-squared' linear brute force rendering function calls

  for each model
    for each triangle
	  set normal
	  set color for first vertice
	  set position for first vertice

	  set color for second vertice
	  set position for second vertice

	  set color for third vertice
	  set position for third vertice
*/
GLvoid Viewport::drawModel() {
  set <Triangle *> triangles;
  set <Triangle *>::iterator tri;
  Vertex * v0, * v1, * v2;
  Normal * norm;
  Color * c;

  for (GLint i = 0; i != scene->numModels(); i++) {
    triangles = *(scene->m(i)->getTriangleList());

    for (tri = triangles.begin(); tri != triangles.end(); tri++) {
      v0 = (*tri)->v0();
      v1 = (*tri)->v1();
      v2 = (*tri)->v2();

      norm = (*tri)->getNormal();
      glNormal3f(norm->x(), norm->y(), norm->z());

      c = v0->getColor();
      glColor4f(c->r(), c->g(), c->b(), c->a());
      glVertex3f(v0->x(), v0->y(), v0->z());

      c = v1->getColor();
      glColor4f(c->r(), c->g(), c->b(), c->a());
      glVertex3f(v1->x(), v1->y(), v1->z());

      c = v2->getColor();
      glColor4f(c->r(), c->g(), c->b(), c->a());
      glVertex3f(v2->x(), v2->y(), v2->z());
    }
  }
}


GLvoid Viewport::drawWireframe() {
  set <Triangle *> triangles;
  set <Triangle *>::iterator tri;
  Vertex * v0, * v1, * v2;
  Normal * norm;

  if (draw_triangles) glColor4f(0.0f, 0.0f, 0.0f, 1.0f);  // set the color to black
  else glColor4f(0.5f, 0.6f, 0.5f, 1.0f);  // set the color to white

  for (GLint i = 0; i != scene->numModels(); i++) {
    triangles = *(scene->m(i)->getTriangleList());

    for (tri = triangles.begin(); tri != triangles.end(); tri++) {
      v0 = (*tri)->v0();
      v1 = (*tri)->v1();
      v2 = (*tri)->v2();

      norm = (*tri)->getNormal();

      glBegin(GL_LINES);
        glVertex3f(v0->x(), v0->y(), v0->z());
        glVertex3f(v1->x(), v1->y(), v1->z());
      glEnd();

      glBegin(GL_LINES);
        glVertex3f(v0->x(), v0->y(), v0->z());
        glVertex3f(v2->x(), v2->y(), v2->z());
      glEnd();

      glBegin(GL_LINES);
        glVertex3f(v1->x(), v1->y(), v1->z());
        glVertex3f(v2->x(), v2->y(), v2->z());
      glEnd();
    }
  }
}


// ------
// Routine which actually do the drawing

GLvoid Viewport::render() {
  // Enables Smooth Color Shading if smooth is true
  if (smooth) glShadeModel(GL_SMOOTH);
  else glShadeModel(GL_FLAT);

  if (lighting) {
    if ((draw_points  || draw_edges) && !draw_triangles)
      glDisable(GL_LIGHTING);
	else
      glEnable(GL_LIGHTING);
  }
  else glDisable(GL_LIGHTING);

  /*
    Do we really need to do this in here or can we do it only when a reshape call happens???
  */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, frustrumDepth);
  glViewport(startx, starty, width, height);

  // Need to manipulate the ModelView matrix to move our model around.
  glMatrixMode(GL_MODELVIEW);

  // Reset to 0,0,0; no rotation, no scaling.
  glLoadIdentity();

  /*
    If we translate before we rotate we can change the model
    else we rotate the camera eye
  */

  glPushMatrix();

  if (control == MODEL_CONTROL) {    // rotate model
    glTranslatef(X_Off, Y_Off, Z_Off);

    glTranslatef(center.x(), center.y(), center.z());
    // Rotate the calculated amount.
    glRotatef(X_Rot, 1.0f, 0.0f, 0.0f);
    glRotatef(Y_Rot, 0.0f, 1.0f, 0.0f);
    glRotatef(Z_Rot, 0.0f, 0.0f, 1.0f);
    glTranslatef(-center.x(), -center.y(), -center.z());
  }
  else {
    // Rotate the calculated amount.
    glRotatef(X_Rot, 1.0f, 0.0f, 0.0f);
    glRotatef(Y_Rot, 0.0f, 1.0f, 0.0f);
    glRotatef(Z_Rot, 0.0f, 0.0f, 1.0f);

    glTranslatef(X_Off, Y_Off, Z_Off);
  }

  if (draw_edges) {
    // Begin drawing lines
    drawWireframe();
  }
  if (draw_points) {
    // Begin drawing points
    glBegin(GL_POINTS);
#ifdef USE_DISPLAY_LISTS
    glCallList(MODEL_LIST);
#else
    drawModel();
#endif
    glEnd();
  }
  if (draw_triangles) {
    // Begin drawing triangles
    glBegin(GL_TRIANGLES);
#ifdef USE_DISPLAY_LISTS
    glCallList(MODEL_LIST);
#else
    drawModel();
#endif
    glEnd();
  }

  /*
    So far Viewport can only draw normals if it uses display lists
    This can be a problem. To solve this problem make another vertex array
    containing all of the data below.
  */
  if (normals) {
    glBegin(GL_LINES);
    glDisable(GL_LIGHTING);
    glColor3f(1.0f, 0.0f, 0.0f);
    drawNormals();
    glEnd();
  }
  else if (vert_normals) {
    glBegin(GL_LINES);
    glDisable(GL_LIGHTING);
    glColor3f(1.0f, 0.0f, 0.0f);
    drawVertNormals();
    glEnd();
  }

  // Done drawing model
  glPopMatrix();
}


GLint Viewport::getStartX() { return startx; }
GLint Viewport::getStartY() { return starty; }
GLint Viewport::getWidth() { return width; }
GLint Viewport::getHeight() { return height; }

GLint Viewport::getViewportID() { return viewportID; }
Scene * Viewport::getScene() { return scene; }

bool Viewport::setStartX(GLint x) { startx = x; return true; }
bool Viewport::setStartY(GLint y) { starty = y; return true; }
bool Viewport::setWidth(GLint w) { width = w; return true; }
bool Viewport::setHeight(GLint h) { height = h; return true; }

bool Viewport::setViewportID(GLint v) { viewportID = v; return true; }
bool Viewport::setScene(Scene * s) { scene = s; return true; }
bool Viewport::setFilename(string file) { filename = file; return true; }

string Viewport::getFilename() { return filename; }

GLvoid Viewport::reset() { loaded = load(filename); startingView(); }
bool Viewport::fileOk() { return loaded; }

GLvoid Viewport::reportStatistics() {
  for (GLint i = 0; i != scene->numModels(); i++) {
    cout << "\n\nViewport "<< viewportID << " Statistics:" << endl;
    cout << "\n\tModel:\t" << i << endl;
    cout << "\tVertices:\t" << scene->m(i)->getVertexList()->size() << endl;
    cout << "\tTriangles:\t" << scene->m(i)->getTriangleList()->size() << endl;
    cout << "\tClockwise:\t" << (scene->m(i)->isClockwise()?"true":"false") << endl;
  }
}

GLvoid Viewport::makeSmooth() { smooth = true; }
GLvoid Viewport::makeFlat() { smooth = false; }
GLvoid Viewport::switchLighting() { lighting = !lighting; }

GLvoid Viewport::switchViewPoints() { draw_points = !draw_points; }
GLvoid Viewport::switchViewEdges() { draw_edges = !draw_edges; }
GLvoid Viewport::switchViewTriangles() { draw_triangles = !draw_triangles; }

GLvoid Viewport::switchNormals() {
  vert_normals = false;
  normals = !normals;
}

GLvoid Viewport::switchVertNormals() {
  normals = false;
  vert_normals = !vert_normals;
}

GLvoid Viewport::setCameraControl(GLint c) { control = c; startingView(); }

GLvoid Viewport::decreaseX_Off() { X_Off -= TRANS_STEP; }
GLvoid Viewport::increaseX_Off() { X_Off += TRANS_STEP; }
GLvoid Viewport::decreaseY_Off() { Y_Off -= TRANS_STEP; }
GLvoid Viewport::increaseY_Off() { Y_Off += TRANS_STEP; }
GLvoid Viewport::decreaseZ_Off() { Z_Off -= TRANS_STEP; }
GLvoid Viewport::increaseZ_Off() { Z_Off += TRANS_STEP; }

GLvoid Viewport::mouseX_Rot(float f) { X_Rot -= f; }
GLvoid Viewport::mouseY_Rot(float f) { Y_Rot -= f; }
GLvoid Viewport::mouseZ_Rot(float f) { Z_Rot -= f; }

GLvoid Viewport::decreaseX_Rot() { X_Rot -= ROT_STEP; }
GLvoid Viewport::increaseX_Rot() { X_Rot += ROT_STEP; }
GLvoid Viewport::decreaseY_Rot() { Y_Rot -= ROT_STEP; }
GLvoid Viewport::increaseY_Rot() { Y_Rot += ROT_STEP; }
GLvoid Viewport::decreaseZ_Rot() { Z_Rot -= ROT_STEP; }
GLvoid Viewport::increaseZ_Rot() { Z_Rot += ROT_STEP; }
