// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#ifndef VIEWPORT_H
#define VIEWPORT_H

#define DEFAULT_WIDTH 300
#define DEFAULT_HEIGHT 300
#define MODEL_CONTROL 0x0401
#define EYE_CONTROL 0x0402

#define PERCENTAGE 0x0101
#define NUMBER 0x0102
#define DEFAULT 0x0103

#include "options.h"
#include "VRMLParser.h"
#include "Decimate.h"
#include <iostream>
#include <GL/gl.h>
#include <GL/glu.h>

class Viewport {
  GLint startx;
  GLint starty;
  GLint width;
  GLint height;
  Scene * scene;
  std::string filename;

  GLfloat TRANS_STEP;
  GLfloat ROT_STEP;

  Point center;
  GLfloat backUp;
  GLint viewportID;

  bool loaded;

  /*
    Transformation and rotation settings
  */
  GLfloat normalLength;
  GLfloat frustrumDepth;
  GLfloat X_Rot;
  GLfloat Y_Rot;
  GLfloat Z_Rot;

  GLfloat X_Off;
  GLfloat Y_Off;
  GLfloat Z_Off;

  /*
    Viewport drawing settings
  */
  bool lighting;
  bool smooth;
  bool normals;
  bool vert_normals;

  bool draw_points;
  bool draw_edges;
  bool draw_triangles;

  GLint control;

  GLvoid startingView();
  GLvoid defaultInit();

  /*
    For the display list
  */
#ifdef USE_DISPLAY_LISTS
  GLuint partList;
  GLuint MODEL_LIST;
#endif

  /*
    For normal functions
  */
  Point getCenter(Triangle * t);
  GLvoid drawNormals();
  GLvoid drawVertNormals();
  GLvoid drawModel();
  GLvoid drawWireframe();
  GLfloat findDistance(Vertex * v, std::set <Vertex *> * neighbors);

 public:

  GLint getMax(GLfloat d, GLint TYPE);
  //! Constructor
  Viewport(
    std::string file, //!< The filename for the VRML file of the Scene to display in the Viewport
    GLint id, //!< The ID of this Viewport
    bool GUI = true
  );

  //! Destructor
  ~Viewport();

  //! Save the Scene in the Viewport to file <file>
  bool save(std::string file = "NONE_SPECIFIED");

  //! Load the Scene from file <file> into the Viewport
  bool load(std::string file);

  //! Reload the Scene from the file currently viewed
  bool reload();

  //! Is the file a valid VRML file?
  bool fileOk();

  //! Render a frame in the Viewport
  GLvoid render();

  //! Reset the size of the Viewport
  bool resetSize(
    GLint x, //!< The x coordinate of the center of the view frustum
    GLint y, //!< The y coordinate of the center of the view frustum
    GLint w, //!< The width of the view frustum
    GLint h  //!< The height of the view frustum
  );

  //! Decimate the Scene
  bool decimate(
    GLfloat d, //!< The threshhold parameter (how far to decimate)
    GLint TYPE //!< Type of decimation: by PERCENTAGE, or a certain NUMBER of Verticies to remove
  );

  GLint getStartX();     //!< Get the starting x coordinate of the view frustum
  GLint getStartY();     //!< Get the starting y coordinate of the view frustum
  GLint getWidth();      //!< Get the starting width of the view frustum
  GLint getHeight();     //!< Get the starting height of the view frustum
  GLint getViewportID(); //!< Get the ID of this Viewport
  Scene * getScene();    //!< Get the Scene in this Viewport

  bool setStartX(GLint x); //!< Set the starting x coordinate of the view frustum
  bool setStartY(GLint y); //!< Set the starting y coordinate of the view frustum
  bool setWidth(GLint w);  //!< Set the starting width of the view frustum
  bool setHeight(GLint h); //!< Set the starting height of the view frustum

  bool setViewportID(GLint v);        //!< Set the Viewport's ID
  bool setScene(Scene * s);           //!< Set the Scene in this Viewport
  bool setFilename(std::string file); //!< Set the filename for the Scene

  std::string getFilename(); //!< Get the filename for the Scene

  GLvoid reset(); //!< Reload the Scene from the file and reset the view

  GLvoid reportStatistics();  //!< Displays current Model information
  GLvoid makeSmooth();        //!< Set to render Models with smooth surfaces
  GLvoid makeFlat();          //!< Set to render Models with flat surfaces
  GLvoid switchLighting();    //!< Toggle the type of lighting
  GLvoid switchNormals();     //!< Toggle display of Triangle Normals on and off
  GLvoid switchVertNormals(); //!< Toggle display of Vertex Normals on and off

  GLvoid switchViewPoints();    //!< Toggle display of Vertices (points)
  GLvoid switchViewEdges();     //!< Toggle display of edges (wireframe)
  GLvoid switchViewTriangles(); //!< Toggle display of Triangles (faces)

  GLvoid setCameraControl(GLint c); //!< Change to Eye or Model control

  GLvoid decreaseX_Off(); //!< Move the camera left
  GLvoid increaseX_Off(); //!< Move the camera right
  GLvoid decreaseY_Off(); //!< Move the camera up
  GLvoid increaseY_Off(); //!< Move the camera down
  GLvoid decreaseZ_Off(); //!< Move the camera back
  GLvoid increaseZ_Off(); //!< Move the camera forward

  GLvoid mouseX_Rot(GLfloat f); //!< Tilt camera forward or back (around X axis)
  GLvoid mouseY_Rot(GLfloat f); //!< Turn camera left or right (around the y axis)
  GLvoid mouseZ_Rot(GLfloat f); //!< Rotate the camera left or right (around the z axis)

  GLvoid decreaseX_Rot(); //!< Tilt the camera back (around x axis)
  GLvoid increaseX_Rot(); //!< Tilt the camera forward (around x axis)
  GLvoid decreaseY_Rot(); //!< Turn camera left (around y axis)
  GLvoid increaseY_Rot(); //!< Turn camera right (around y axis)
  GLvoid decreaseZ_Rot(); //!< Rotate camera left (around z axis)
  GLvoid increaseZ_Rot(); //!< Rotate camera right (around z axis)
};

#endif
