// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "ViewportController.h"

using namespace std;

ViewportController::ViewportController(int maxNumViewports, int w, int h) {
  width = w;
  height = h;
  viewportList.reserve(maxNumViewports);
}


ViewportController::~ViewportController() {
  for (vector <Viewport>::iterator iter = viewportList.begin();
       iter != viewportList.end();
       iter++) {
    viewportList.erase(iter);
  }
}


bool ViewportController::resizeScene(int Width, int Height) {
  int numView = numViewports();
  int prevX, prevY;
  int view = 0;
  int cols;
  int rows;
  double n = sqrt((double)numView);

  // Let's not core dump, no matter what.
  if (Height == 0) { Height = 1; }

  /*
    This will get the number of rows and columns, by detecting a perfect square
  */
  if ((n - (int) n) > -0.0001 && (n - (int) n) < 0.0001) {
    cols = (int) n;
    rows = (int) n;
  }
  else if (n > (((int) n) + 0.5)) {    // Adding to the bottom row
    cols = ((int) n) + 1;
    rows = ((int) n) + 1;
  }
  else {    // Adding to the rightside column
    cols = ((int) n) + 1;
    rows = (int) n;
  }


  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "n = " << n
	 << "; cols = " << cols
	 << "; rows = " << rows
	 << "; Width = " << Width
	 << "; Height = " << Height
	 << endl;

  int stepW = Width / cols;
  int stepH = Height / rows;
  bool new_row = true;

  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "ViewportController::Screen: Width = " << Width << "; Height = " << Height << "\n";

  for (int i = 0; i != rows; i++) {
    if (view == 0) {
      prevX = 0;
      prevY = 0;
      getViewport(view)->setStartX(0);
      getViewport(view)->setStartY(0);
    }
    else {    // Add to the height side, width doesn't change
      prevX = getViewport(view-cols)->getStartX();
      prevY = getViewport(view-cols)->getStartY() + stepH;
    }

    new_row = true;
    for (int j = 0; j != cols; j++) {
      if (!new_row) {    // Add to the width side, height doesn't change
		prevX = getViewport(view-1)->getStartX() + stepW;
		prevY = getViewport(view-1)->getStartY();
      }
      new_row = false;

      getViewport(view)->resetSize(prevX, prevY, stepW, stepH);

      if (VIEWPORT_CONTROLLER_DEBUG)
		cout << "ViewportController::Viewport #" << view
				  << ": x = " << getViewport(view)->getStartX()
				  << "; y = " << getViewport(view)->getStartY()
				  << "; width = " << getViewport(view)->getWidth()
				  << "; height = " << getViewport(view)->getHeight() << "\n";

      if(++view == numView) break;    // this should be safe since it will exit in the next outside for loop call
    }
  }

  return true;
}


/*
  creates a new Viewport and adds it to the viewportList, returns true
  if the viewport was created succesfully

  pre-condition: numViewports() != MAX_VIEWPORTS
*/
bool ViewportController::createViewport(string file) {
  if (file == "NONE_SPECIFIED") {
    file = filename;
  }
  else
    filename = file;

  int numView;
  int cols;
  int rows;
  double n;

  /*
    Is this the first viewport
  */
  if (numViewports() == 0) {
    currentViewport = new Viewport(filename, 0);

    if(!currentViewport->fileOk())
      return false;

    viewportList.insert(viewportList.end(), *currentViewport);
    return true;
  }

  // We add one, because we are adding a viewport
  numView = numViewports() + 1;

  // Checks to see if numView is a perfect square
  n = sqrt((double)numView);
  if ((n - (int) n) > -0.0001 && (n - (int) n) < 0.0001) {
    cols = (int) n;
    rows = (int) n;
  }
  else if (n > (((int) n) + 0.5)) {    // Adding to the bottom row
    cols = ((int) n) + 1;
    rows = ((int) n) + 1;
  }
  else {    // Adding to the rightside column
    cols = ((int) n) + 1;
    rows = (int) n;
  }

  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "ViewportController::ADD_VIEWPORT: viewportID = " << numView - 1 << endl;

  // Gets the common viewport width and height
  width = currentViewport->getWidth() * cols;
  height = currentViewport->getHeight() * rows;

  currentViewport = new Viewport(filename, numView - 1);

  if (currentViewport->fileOk()) {
    viewportList.insert(viewportList.end(), *currentViewport);
    return true;
  }
  else
    return false;
}

/*
  pre-condition: numViewports() != 0
*/
bool ViewportController::deleteViewport() {
  int numView;
  int cols;
  int rows;
  double n;

  // We subtract one, because we are removing a viewport
  numView = numViewports() - 1;

  // Checks to see if numView is a perfect square
  n = sqrt((double)numView);
  if ((n - (int) n) > -0.0001 && (n - (int) n) < 0.0001) {
    cols = (int) n;
    rows = (int) n;
  }
  else if (n > (((int) n) + 0.5)) {    // Removing from the bottom row
    cols = ((int) n) + 1;
    rows = ((int) n) + 1;
  }
  else {    // Removing from the rightside column
    cols = ((int) n) + 1;
    rows = (int) n;
  }

  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "ViewportController::REMOVE_VIEWPORT: viewportID = " << currentViewport->getViewportID() << endl;

  width = currentViewport->getWidth() * cols;
  height = currentViewport->getHeight() * rows;

  int viewportID = currentViewport->getViewportID();
  currentViewport = remove(currentViewport);

  fixViewportIDs(viewportID);

  return true;
}


void ViewportController::fixViewportIDs(int viewportID) {
  for (vector <Viewport>::iterator iter = viewportList.begin() + viewportID; iter != viewportList.end(); iter++)
    iter->setViewportID(iter->getViewportID() - 1);
}


Viewport * ViewportController::remove(Viewport * v) {
  /*
    Find the viewport first
  */
  for (vector <Viewport>::iterator iter = viewportList.begin();
       iter != viewportList.end();
       iter++) {
    if (&(*iter) == v) {
      viewportList.erase(iter);
      break;
    }
  }

  return &(*viewportList.begin());
}


void ViewportController::renderScene() {
  for (vector <Viewport>::iterator iter = viewportList.begin(); iter != viewportList.end(); iter++)
    iter->render();
}


int ViewportController::numViewports() { return (int)(viewportList.size()); }


bool ViewportController::selectCurrentViewport(int x, int y) {
  vector <Viewport>::iterator iter;

  for (iter = viewportList.begin();
       iter != viewportList.end();
       iter++) {
    if((iter->getStartX() + iter->getWidth() ) > x) {
      if(( iter->getStartY() + iter->getHeight() ) > y) {
        currentViewport = &( *iter );

	if(VIEWPORT_CONTROLLER_DEBUG)
	  cout << "ViewportController::Selected viewport #" << iter->getViewportID() << endl;

	return true;
      }
    }
  }

  currentViewport = &(*viewportList.begin());
  return false;
}


Viewport * ViewportController::getViewport(int viewportID) {
  vector <Viewport>::iterator iter;

  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "ViewportController::viewportID = " << viewportID << endl;

  for (iter = viewportList.begin();
       iter != viewportList.end();
       iter++) {

    if (iter->getViewportID() == viewportID) {
      if (VIEWPORT_CONTROLLER_DEBUG)
		cout << "ViewportController::found viewportID = " << viewportID << endl;

      return &(*iter);
    }
  }

  return (&(*viewportList.begin()));
}


Scene * ViewportController::getScene(int viewportID) {
  if (VIEWPORT_CONTROLLER_DEBUG)
    cout << "ViewportController::getScene: viewportID = " << viewportID << endl;

  return (viewportList[viewportID].getScene());
}

string ViewportController::getFilename() { return filename; }

int ViewportController::getWidth() { return width; }
int ViewportController::getHeight() { return height; }
bool ViewportController::save() { return currentViewport->save(); }
bool ViewportController::save(string file) { return currentViewport->save(file); }
bool ViewportController::decimateEvent(GLfloat f, GLint TYPE) { return currentViewport->decimate(f, TYPE); }

void ViewportController::resetViewport() { currentViewport->reset(); }

void ViewportController::reportStatistics() { currentViewport->reportStatistics(); }

void ViewportController::makeSmooth() { currentViewport->makeSmooth(); }
void ViewportController::makeFlat() { currentViewport->makeFlat(); }
void ViewportController::switchViewPoints() { currentViewport->switchViewPoints(); }
void ViewportController::switchViewEdges() { currentViewport->switchViewEdges(); }
void ViewportController::switchViewTriangles() { currentViewport->switchViewTriangles(); }
void ViewportController::switchLighting() { currentViewport->switchLighting(); }
void ViewportController::switchNormals() { currentViewport->switchNormals(); }
void ViewportController::switchVertNormals() { currentViewport->switchVertNormals(); }
void ViewportController::setCameraControl(int control) { currentViewport->setCameraControl(control); }

void ViewportController::decreaseX_Off() { currentViewport->decreaseX_Off(); }
void ViewportController::increaseX_Off() { currentViewport->increaseX_Off(); }
void ViewportController::decreaseY_Off() { currentViewport->decreaseY_Off(); }
void ViewportController::increaseY_Off() { currentViewport->increaseY_Off(); }
void ViewportController::decreaseZ_Off() { currentViewport->decreaseZ_Off(); }
void ViewportController::increaseZ_Off() { currentViewport->increaseZ_Off(); }

void ViewportController::mouseX_Rot(float f) { currentViewport->mouseX_Rot(f); }
void ViewportController::mouseY_Rot(float f) { currentViewport->mouseY_Rot(f); }
void ViewportController::mouseZ_Rot(float f) { currentViewport->mouseZ_Rot(f); }

void ViewportController::decreaseX_Rot() { currentViewport->decreaseX_Rot(); }
void ViewportController::increaseX_Rot() { currentViewport->increaseX_Rot(); }
void ViewportController::decreaseY_Rot() { currentViewport->decreaseY_Rot(); }
void ViewportController::increaseY_Rot() { currentViewport->increaseY_Rot(); }
void ViewportController::decreaseZ_Rot() { currentViewport->decreaseZ_Rot(); }
void ViewportController::increaseZ_Rot() { currentViewport->increaseZ_Rot(); }
