// Mesh Algorithms: a library of algorithms to manipulate 3D meshes
// Copyright (C) 2001 Tim Garthwaite and Jason Reposa

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
/*
  Some code in this file was borrowed from http://oreilly.linux.com/
  Specifically the callback functions.

  NOTES:
  The y value sent to HandleMouseEvent is opposite of what openGL wants.
  i.e. if you touch a point in a window sized (width = 600, height = 300)
  and glut returns say (250,100), while the correct (x,y) that openGL
  returns is (250,200). So in order to correct the problem in
  HandleMouseEvent we say openGLY = (windowHeight - glutY)

  We have successfully added up to six (6) viewports. We can only
  assume that all viewport additions will also be scaled correctly.

  TODO:
  Add text info for each viewport
*/

#include "main.h"

using namespace std;

/*
  Options: (SUBJECT TO CHANGE, ALWAYS!)
  algo <filename> <action> <action-options>

  -h prints this message
  --help prints this message

  <filename> MANDATORY

  <action> MANDATORY
  gui
  nogui

  <nogui-options> (invoked by the "nogui" action)
  -o <filename> the name of the VRML file to save as DEFAULT: out.wrl
  One of these needs to be specified:
  -p <num> percent to decimate to
  -v <num> vertex count to decimate to

  A typical program exectution would be:
  algo model.wrl

  Or an example that does not use the gui would be:
  algo model.wrl nogui -p 100 -o new_chess.wrl
*/

/*!
  Prints the usage as defined above.
*/
void printusage() {
  cout << "\nUsage:\n\n";
  cout << "algo <filename> [action] [action-options]\n";
  cout << "\t-h prints this message\n";
  cout << "\t--help prints this message\n\n";
  cout << "<filename> The VRML file to open\n\n";

  cout << "[action]\n";
  cout << "\tgui DEFAULT\n";
  cout << "\tnogui\n\n";

  cout << "[nogui-options] (options for the \"nogui\" action)\n";
  cout << "\t-o filename \tthe name of the VRML file to save as DEFAULT: out.wrl\n";
  cout << "\tOne of the following are MANDATORY:\n";
  cout << "\t-p number\tpercent to decimate to\n";
  cout << "\t-v number\tvertex count to decimate to\n\n";

  cout << "Typical usage would be:\n";
  cout << "\talgo model.wrl\n\n";

  cout << "Typical usage for nogui would be:\n";
  cout << "\talgo model.wrl nogui -p 100 -o output.wrl\n\n";
}


/*!
  Prints the keyboard controls.
*/
void printcontrols() {
  cout << "\nKeyboard Controls:\n\n";
  cout << "Action:\t\t\t\t\t\t\tButtons\n";
  cout << "-------------------------------------\t\t\t----------------\n";
  cout << "open file (not working):\t\t\t\to, O, F2\n";
  cout << "save - overwrite file (not vorking):\t\t\tF3\n";
  cout << "save as dialog (not working)\t\t\t\ta, A, F4\n";
  cout << "reset viewport:\t\t\t\t\t\tr, R\n";
  
  cout << "eye control:\t\t\t\t\t\te, E\n";
  cout << "model control:\t\t\t\t\t\tm, M\n";
  cout << "render smooth shading:\t\t\t\t\ts, S\n";
  cout << "render flat shading:\t\t\t\t\tf, F\n";

  cout << "toggle point rendering:\t\t\t\t\tp, P\n";
  cout << "toggle wireframe rendering:\t\t\t\tw, W\n";
  cout << "toggle triangle rendering:\t\t\t\tt, T\n";
  cout << "toggle lighting rendering:\t\t\t\tl, L\n";
  cout << "toggle normal rendering from triangle:\t\t\tn, N\n";
  cout << "toggle normal rendering from vertice:\t\t\tv, V\n";

  cout << "decrease z rotation:\t\t\t\t\tDELETE\n";
  cout << "increase z rotation:\t\t\t\t\tEND\n";
  cout << "move model up:\t\t\t\t\t\tPAGE UP\n";
  cout << "move model down:\t\t\t\t\tPAGE DOWN\n";
  cout << "move model left:\t\t\t\t\tLEFT\n";
  cout << "move model right:\t\t\t\t\tRIGHT\n";
  cout << "move model away from viewer:\t\t\t\tUP\n";
  cout << "move model towards viewer:\t\t\t\tDOWN\n";

  cout << "quick surface decimate:\t\t\t\t\td, D, F1\n";
  cout << "remove 1, 2, 3 vertices:\t\t\t\t1, 2, 3\n";
  cout << "quit:\t\t\t\t\t\t\tq, Q, ESC\n\n";

  cout << "\nUse the mouse to control the view of the model.\n";
  cout << "\nRight click to open a menu for further control options.\n\n";
}


/*!
  Called to exit program cleanly.
  This operation is called when the system determines it is necessary
  to exit the program. Reasons for exiting are user decides to quit
  the program, the user did not specify the right command line options,
  or the file that was specified on the command line is incorrect.
*/
void shutDownEvent(int exitType = ERR_CLEAN_EXIT) {
  switch (exitType) {
  case ERR_CLEAN_EXIT:
    glutDestroyWindow(windowID);    // Just in case there is a window open.
    exit (ERR_CLEAN_EXIT);
    break;

  case ERR_PRINT_USAGE:
    printusage();    // Prints the usage to stdout
    exit (ERR_PRINT_USAGE);
    break;

  case ERR_FILE_NOT_FOUND:
    exit (ERR_FILE_NOT_FOUND);
    break;

  case ERR_CONTINUE:
    break;

  default:
    glutDestroyWindow(windowID);    // Just in case there is a window open.
    exit (ERR_UNKNOWN);
    break;
  }
}


/*!
  Print error message, exit cleanly.
  This is a helper operation. It provides the system a way to specify
  a location, message, and if the system needs to exit.
*/
void error(string location, string message, int exitType = ERR_CONTINUE) {
  cerr << "\n\n" << location << ": " << message << endl;

  shutDownEvent(exitType);
}


/*!
  Save Scene to file.
  This operation is called when the user wants to save the viewport's
  scene. We are only expecting to save in VRML format.
*/
void saveEvent(int saveStyle) {
  string file;

  switch (saveStyle) {
  case QUICK_SAVE:
    /*
      Assumes that the filename has already been specified
    */
    if (!viewports->save(outputfile))
      error("saveEvent", "File did not write correctly.");
    break;

  case DIALOG_SAVE:
    /*
      Get the filename:
      file = saveDialog();
    */
    file = outputfile;

    if (!viewports->save(file))
      error("saveEvent", "GLUT does not support save as dialog boxes.");
    else
      outputfile = file;
    break;

  default:    // Should NEVER reach this point.
    error("saveEvent", "Bad saving style. Impossible!!!");
    break;
  }
}

/*!
  Open file event.
  Event callback for when user chooses to load a file while
  program is running. Also used if filename is given on the
  command line.
*/
void openEvent(string s = "PROMPTFORFILENAME") {
  if (s == "PROMPTFORFILENAME")
    error("openEvent", "GLUT does not support open dialog boxes.");
  else
    if (!viewports->createViewport(s)) {
      error("openEvent", "file - " + s + " is not a valid filename.", ERR_FILE_NOT_FOUND);
    }
}

/*!
  OpenGL callback for window resize.
  This is the operation that is specified in glutReshapeFunc(). This
  operation is a callback routine executed whenever our window is resized.
*/
void cbResizeScene(int Width, int Height) {
  // Clear the color and depth buffers.
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  viewports->resizeScene(Width, Height);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  windowWidth  = Width;
  windowHeight = Height;

  glutPostRedisplay();
}


/*!
  OpenGL callback for render event.
  This is the operation that is specified in glutDisplayFunc(). This
  operation is a callback routine executed whenever the window needs
  to be rendered. It is Glut based so it is much simpler and cleaner.
*/
void cbRenderScene() {
  // Clear the color and depth buffers.
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /*
    Render the entire scene
  */
  viewports->renderScene();

  /*
    Draw a rectangle around the selected (current) viewport
  */

  // Swap buffers when done
  glutSwapBuffers();
}

/*!
  OpenGL callback for a keypress event.
  This is the operation that is specified in glutKeyboardFunc(). This
  operation is a callback routine executed whenever a normal key is
  pressed.
*/
void cbKeyPressed(unsigned char key, int x, int y) {
  //! Switch on the key pressed.
  switch (key) {
  case 113: case 81: case 27: //!< q, Q, ESC
    shutDownEvent();
    break;

  case 63: //!< ?
    viewports->reportStatistics();
    break;

  case 101: case 69: //!< e, E
    viewports->setCameraControl(EYE_CONTROL);
    break;

  case 109: case 77: //!< m, M
    viewports->setCameraControl(MODEL_CONTROL);
    break;

  case 114: case 82: //!< r, R
    viewports->resetViewport();
    break;

  case 100: case 68: //!< d, D
    viewports->decimateEvent(1.0f, PERCENTAGE);
    break;

  case 111: case 79: //!< o, O
    openEvent();
    break;

  case 115: case 83: //!< s, S
    viewports->makeSmooth();
    break;

  case 102: case 70: //!< f, F
    viewports->makeFlat();
    break;

  case 112: case 80: //!< p, P
    viewports->switchViewPoints();
    break;

  case 119: case 87: //!< w, W
    viewports->switchViewEdges();
    break;

  case 116: case 84: //!< t, T
    viewports->switchViewTriangles();
    break;

  case 108: case 76: //!< l, L
    viewports->switchLighting();
    break;

  case 110: case 78: //!< n, N
    viewports->switchNormals();
    break;

  case 118: case 86: //!< v, V
    viewports->switchVertNormals();
    break;

  case 97: case 65: //!< a, A
    saveEvent(DIALOG_SAVE);
    break;

  case 127:
    viewports->increaseZ_Rot();
    break;

  case 49:
    viewports->decimateEvent(1.0f, NUMBER);
    break;

  case 50:
    viewports->decimateEvent(2.0f, NUMBER);
    break;

  case 51:
    viewports->decimateEvent(3.0f, NUMBER);
    break;

  default: //!< The key pressed is not defined
    error("cbKeyPressed", "No action for key");
    break;
  }

  glutPostRedisplay();
}

/*!
  OpenGL callback for special keypress event.
  This is the operation that is specified in glutSpecialFunc(). This
  operation is a callback routine executed whenever a special key is
  pressed.
*/
void cbSpecialKeyPressed(int key, int x, int y) {
  //! Switch on the key pressed.
  switch (key) {
  case GLUT_KEY_PAGE_UP:
    viewports->increaseY_Off();
    break;

  case GLUT_KEY_PAGE_DOWN:
    viewports->decreaseY_Off();
    break;

  case GLUT_KEY_UP:
    viewports->decreaseZ_Off();
    break;

  case GLUT_KEY_DOWN:
    viewports->increaseZ_Off();
    break;

  case GLUT_KEY_LEFT:
    viewports->decreaseX_Off();
    break;

  case GLUT_KEY_RIGHT:
    viewports->increaseX_Off();
    break;

  case GLUT_KEY_END:
    viewports->decreaseZ_Rot();
    break;

  case GLUT_KEY_F1: //!< F1
    viewports->decimateEvent(1.0f, PERCENTAGE);
    break;

  case GLUT_KEY_F2: //!< F2
    openEvent();
    break;

  case GLUT_KEY_F3: //!< F3
    saveEvent(QUICK_SAVE);
    break;

  case GLUT_KEY_F4: //!< F4
    saveEvent(DIALOG_SAVE);
    break;

  default: //!< The key pressed is not defined
    error("cbSpecialKeyPressed", "No action for special key");
    break;
  }

  glutPostRedisplay();
}

/*!
  OpenGL callback for clicked mouse motion event.
  This is the operation that is specified in glutMotionFunc(). This
  operation is a callback routine executed whenever the mouse is moved.
*/
void HandleMotionEvent(int x, int y) {
  viewports->mouseY_Rot((float)mouseX - x);
  mouseX = x;

  viewports->mouseX_Rot((float)mouseY - y);
  mouseY = y;

  glutPostRedisplay();
}

/*!
  OpenGL callback for mouse click event.
  This is the operation that is specified in glutMouseFunc(). This
  operation is a callback routine executed whenever the left or middle
  mouse button is pressed.
*/
void HandleMouseEvent(int button, int state, int x, int y) {
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    mouseX = x;
    mouseY = y;

    // Also get the currentViewport and redraw
    viewports->selectCurrentViewport(x, (windowHeight - y));
  }
}

/*!
  OpenGL callback for right mouse button click.
  This is the operation that is specified in glutCreateMenu(). This
  operation is a callback routine executed whenever the right mouse
  button is pressed.
*/
void HandleMenuEvent(int menuitem) {
  switch (menuitem) {
  case MODEL_CONTROL:
    viewports->setCameraControl(MODEL_CONTROL);
    break;

  case EYE_CONTROL:
    viewports->setCameraControl(EYE_CONTROL);
    break;

  case SHOW_POINTS:
    viewports->switchViewPoints();
    break;

  case SHOW_EDGES:
    viewports->switchViewEdges();
    break;

  case SHOW_TRIANGLES:
    viewports->switchViewTriangles();
    break;

  case SHOW_LIGHTING:
    viewports->switchLighting();
    break;

  case SHOW_NORMALS:
    viewports->switchNormals();
    break;

  case SHOW_VERT_NORMALS:
    viewports->switchVertNormals();
    break;

  case ADD_VIEWPORT:
    if (viewports->numViewports() == MAX_VIEWPORTS) break;

    viewports->createViewport();

    cbResizeScene(viewports->getWidth(), viewports->getHeight());
    glutReshapeWindow(windowWidth, windowHeight);
    glutPostRedisplay();
    glutPostRedisplay();
    break;

  case REMOVE_VIEWPORT:
    if (viewports->numViewports() == 1) break;

    viewports->deleteViewport();

    cbResizeScene(viewports->getWidth(), viewports->getHeight());
    glutReshapeWindow(windowWidth, windowHeight);
    glutPostRedisplay();
    glutPostRedisplay();
    break;

  case DEC_20_PER:
    viewports->decimateEvent(0.2f, PERCENTAGE);
    break;

  case DEC_40_PER:
    viewports->decimateEvent(0.4f, PERCENTAGE);
    break;

  case DEC_60_PER:
    viewports->decimateEvent(0.6f, PERCENTAGE);
    break;

  case DEC_80_PER:
    viewports->decimateEvent(0.8f, PERCENTAGE);
    break;

  case DEC_90_PER:
    viewports->decimateEvent(0.9f, PERCENTAGE);
    break;

  case DEC_ONE_TENTH:
    viewports->decimateEvent(0.001f, PERCENTAGE);
    break;
  case DEC_ONE:
    viewports->decimateEvent(0.01f, PERCENTAGE);
    break;
  case DEC_TWO_HALF:
    viewports->decimateEvent(0.025f, PERCENTAGE);
    break;
  case DEC_FIVE:
    viewports->decimateEvent(0.05f, PERCENTAGE);
    break;
  case DEC_TEN:
    viewports->decimateEvent(0.1f, PERCENTAGE);
    break;

  case EXIT:
    shutDownEvent();
    break;

  default:    // Should NEVER reach this point
    error("HandleMenuEvent", "Bad Menu Item Selected. Impossible!!!");
    break;
  }

  glutPostRedisplay();
}

/*!
  Defines the global variables.
*/
void ourInit() {
  /*!
    The viewports that render the scene. This controls the viewports of
    the window. There are many events that call member functions of this
    variable.
  */
  viewports = new ViewportController(MAX_VIEWPORTS, DEFAULT_WIDTH, DEFAULT_HEIGHT);

  /*!
    The width of the entire window. This value changes when
    viewports are created or removed.
  */
  windowWidth = DEFAULT_WIDTH;
  /*!
    The height of the entire window. This value changes when
    viewports are created or removed.
  */
  windowHeight = DEFAULT_HEIGHT;

  /*!
    Settings for our light. Currently there are only three defined
    attributes. We may not need these at all.
  */
  Light_Ambient[0] = 0.1f;
  Light_Ambient[1] = 0.1f;
  Light_Ambient[2] = 0.1f;
  Light_Ambient[3] = 1.0f;

  Light_Diffuse[0] = 1.2f;
  Light_Diffuse[1] = 1.2f;
  Light_Diffuse[2] = 1.2f;
  Light_Diffuse[3] = 1.0f;

  Light_Position[0] = 0.0f;
  Light_Position[1] = 0.0f;
  Light_Position[2] = 0.0f;
  Light_Position[3] = 1.0f;

  /*!
    Default settings: makes it less of a pain from the command line
  */
  percent = false;
  counter = false;
  gui = true;
}


/*!
  Pre-GLUT control initialization.
  Does everything needed before losing control to the main OpenGL
  event loop, which in this case is glutMainLoop().
*/
void ourGLInit(int Width, int Height) {
  //! Color to clear color buffer to.
  glClearColor(0.1f, 0.1f, 0.1f, 0.0f);

  //! Depth to clear depth buffer to; type of test.
  glClearDepth(1.0f);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);

  //! A little performance hit, but some say it is worth it.
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  //! Enables us to use different colors with the rendering
  glEnable(GL_COLOR_MATERIAL);

  //! Load up the correct perspective matrix; using a callback directly.
  cbResizeScene(Width, Height);

  //! Set up a light, turn it on.
  glLightfv(GL_LIGHT1, GL_POSITION, Light_Position);
  glLightfv(GL_LIGHT1, GL_AMBIENT,  Light_Ambient);
  glLightfv(GL_LIGHT1, GL_DIFFUSE,  Light_Diffuse);
  glEnable(GL_LIGHT1);
}


/*!
  Main entry point into the system.
*/
int main(int argc, char * argv[]) {
  /*
    For a list of options see printusage() at the beginning of this file.

    Set up some helper variables:
  */
  bool satisfied = false;
  int iter = 1;
  string arg;
  string filename;
  outputfile = "out.wrl";  // the default name of the VRML output file

  ourInit();

  // check to see if the user passed any args.
  if (argc == 1)
    error("main", "No options specified!", ERR_PRINT_USAGE);

  // The filename is mandatory, so we assume that we got it.
  filename = argv[iter];

  // The user specified more than one argument
  if (argc > 2) {
    iter++;
    arg = argv[iter];

    // Is the user looking for help?
    if (arg == "--help" || arg == "-h") {
      error("main", "Help!", ERR_PRINT_USAGE);
    }

    // If not then the user should specify either nogui or gui
    for (; iter != argc; iter++) {
      arg = argv[iter];

      if (arg == "gui")
	break;
      else if (arg == "nogui") {
	gui = false;
	break;
      }
      else
	error("main","Ill formed options for " + (string)argv[0], ERR_PRINT_USAGE);
    }
  }

  // Get action and it's options
  if (!gui) {
    float num = 0.0f;
    for (iter++; iter < argc; iter++) {
      arg = argv[iter];

      // this is not mandatory
      if (arg == "-o") {
        if (argv[++iter]) {
          outputfile = argv[iter];
          satisfied = true;
		}
        else
          error("main", "no filename specified", ERR_PRINT_USAGE);
      }
      // one of these are
      else if (arg == "-p") {
        percent = true;
        if (argv[++iter]) {
          stringstream s;
          s << argv[iter];
          s >> num;
          if (num <= 0 || num > 100) error("main", "percentage is not between 0 and 100.", ERR_PRINT_USAGE);
          if (num > 1) num /= 100;
        }
        else
          error("main", "no percentage specified", ERR_PRINT_USAGE);
      }
      else if (arg == "-v") {
        counter = true;
        if (argv[++iter]) {
          stringstream s;
          s << argv[iter];
          s >> num;
          if (num <= 0) error("main", "count less than or equal to 0.", ERR_PRINT_USAGE);
        }
        else
          error("main", "no count specified", ERR_PRINT_USAGE);
      }
      else { cout << "arg = " << arg << endl;
        error("main",
	      "Ill formed options for action nogui.",
		  ERR_PRINT_USAGE);}
    }

    // decimate to
    // save file as
    // Try to create a viewport based on what was passes as argv[1]
    if (!viewports->createViewport(filename))
      error("main",
	    "Specify a valid filename.",
	    ERR_FILE_NOT_FOUND);

    cout << "\n\nStarting decimation:" << endl;
    if (percent) {
      if (viewports->decimateEvent(num, PERCENTAGE))
        cout << "Decimation successful!" << endl;
      else
        cout << "Decimation unsuccesful: percentage" << endl;
	}
	else if (counter) {  // is there anyway to tell if the num is too large?
      if (viewports->decimateEvent(num, NUMBER))
        cout << "Decimation successful!" << endl;
      else
        cout << "Decimation unsuccesful: number" << endl;
	}
    else  // should never get here
      cout << "Decimation unsuccesful: unknown method" << endl;

    if (viewports->save(outputfile))
      cout << "The decimated scene was saved to " << outputfile << endl;
    else
      cout << "The file was not saved to " << outputfile << endl;
  }

  // If user wants to render the file start GUI.
  else if(gui) {
    glutInit(&argc, argv);

    /*
      GLUT_DEPTH: ???
      GLUT_DOUBLE: Double buffer
      GLUT_RGB: RGB mode no Alpha
    */
    // To see OpenGL drawing, take out the GLUT_DOUBLE request.
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(windowWidth, windowHeight);

    // Open a window 
    windowID = glutCreateWindow( PROGRAM_TITLE );

    // Register the callback function to do the drawing. 
    glutDisplayFunc(&cbRenderScene);

    // It's a good idea to know when our window's resized.
    glutReshapeFunc(&cbResizeScene);

    decPerMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddMenuEntry("remove 20%", DEC_20_PER);
    glutAddMenuEntry("remove 40%", DEC_40_PER);
    glutAddMenuEntry("remove 60%", DEC_60_PER);
    glutAddMenuEntry("remove 80%", DEC_80_PER);
    glutAddMenuEntry("remove 90%", DEC_90_PER);

    decDisMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddMenuEntry("remove 0.1%", DEC_ONE_TENTH);
    glutAddMenuEntry("remove 1%", DEC_ONE);
    glutAddMenuEntry("remove 2.5%", DEC_TWO_HALF);
    glutAddMenuEntry("remove 5%", DEC_FIVE);
    glutAddMenuEntry("remove 10%", DEC_TEN);

    decimationMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddSubMenu("By percentage", decPerMenuID);
    glutAddSubMenu("By smaller percentage", decDisMenuID);

    modelMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddMenuEntry("Lighting On/Off", SHOW_LIGHTING);

    glutAddMenuEntry("Normals On/Off", SHOW_NORMALS);
    glutAddMenuEntry("Vertice Normals On/Off", SHOW_VERT_NORMALS);

    glutAddMenuEntry("Render Points", SHOW_POINTS);
    glutAddMenuEntry("Render Edges ", SHOW_EDGES);
    glutAddMenuEntry("Render Triangles", SHOW_TRIANGLES);

    viewportMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddMenuEntry("Add Viewport", ADD_VIEWPORT);
    glutAddMenuEntry("Remove Viewport", REMOVE_VIEWPORT);

    controlMenuID = glutCreateMenu(HandleMenuEvent);
    glutAddMenuEntry("Model Control", MODEL_CONTROL);
    glutAddMenuEntry("Eye Control", EYE_CONTROL);

    menuID = glutCreateMenu(HandleMenuEvent);
    glutAddSubMenu("Decimation Options", decimationMenuID);
    glutAddSubMenu("Model Options", modelMenuID);
    glutAddSubMenu("Viewport Options", viewportMenuID);
    glutAddSubMenu("Control Options", controlMenuID);
    glutAddMenuEntry("Quit", EXIT);
    glutAttachMenu(GLUT_RIGHT_BUTTON);

    // And let's get some keyboard input.
    glutKeyboardFunc(&cbKeyPressed);
    glutSpecialFunc(&cbSpecialKeyPressed);
    glutMouseFunc(HandleMouseEvent);
    glutMotionFunc(HandleMotionEvent);

    // Try to create a viewport based on what was passes as argv[1]
    if (!viewports->createViewport(filename))
      error("main",
	    "Specify a valid filename.",
	    ERR_FILE_NOT_FOUND);

    // OK, OpenGL's ready to go.  Let's call our own init function.
    ourGLInit(windowWidth, windowHeight);

    // Since it looks like everything went well we display the controls
	printcontrols();
    
	// Pass off control to OpenGL.
    // Above functions are called as appropriate.
    glutMainLoop();
  }

  return 0;
}
